import importlib.metadata

__version__ = importlib.metadata.version("MEArec")

from MEArec.generation_tools import (gen_recordings, gen_spiketrains,
                                     gen_templates)
from MEArec.generators import (RecordingGenerator, SpikeTrainGenerator,
                               TemplateGenerator)
from MEArec.simulate_cells import (calc_extracellular,
                                   calculate_extracellular_potential,
                                   return_bbp_cell, return_bbp_cell_morphology,
                                   run_cell_model, simulate_templates_one_cell)
from MEArec.tools import (available_probes, convert_recording_to_new_version,
                          extract_units_drift_vector,
                          get_default_cell_models_folder, get_default_config,
                          get_default_drift_dict,
                          get_default_recordings_params,
                          get_default_templates_params, get_templates_features,
                          load_dict_from_hdf5, load_recordings,
                          load_recordings_from_file, load_templates,
                          plot_amplitudes, plot_cell_drifts, plot_pca_map,
                          plot_rasters, plot_recordings, plot_templates,
                          plot_waveforms, safe_yaml_load, save_dict_to_hdf5,
                          save_recording_generator, save_recording_to_file,
                          save_template_generator)

from MEArec.drift_tools import generate_drift_dict_from_params



import os
import pprint
import time
from pathlib import Path

import click
import MEAutility as mu
import numpy as np
import yaml
from packaging.version import parse

import MEArec.generation_tools as gt
from MEArec import (get_default_config, save_recording_generator,
                    save_template_generator)
from MEArec.tools import safe_yaml_load


@click.group()
def cli():
    """MEArec: Fast and customizable simulation of extracellular recordings on Multi-Electrode-Arrays"""
    pass


@cli.command()
@click.option(
    "--params",
    "-prm",
    default=None,
    help="path to default_params.yaml (otherwise default default_params are used and some of the parameters"
    "can be overwritten with the following options)",
)
@click.option("--default", is_flag=True, help="shows default values for simulation")
@click.option("--fname", "-fn", default=None, help="template filename")
@click.option("--folder", "-fol", default=None, help="templates output base folder")
@click.option("--cellfolder", "-cf", default=None, help="folder containing bbp cell models")
@click.option(
    "--rot",
    "-r",
    default=None,
    help="possible rotation arguments: norot-xrot-yrot-zrot-physrot-3drot (default=physrot)",
)
@click.option("--probe", "-prb", default=None, help="probe name from available electrodes (default=None)")
@click.option("--n", "-n", default=None, type=int, help="number of observations per cell type (default=50)")
@click.option("--njobs", "-nj", default=0, type=int, help="number of jobs (default n_jobs=n_cpus)")
@click.option("--dt", "-dt", default=None, type=float, help="time period in ms (default=0.03125)")
@click.option("--ncontacts", "-nc", default=None, type=int, help="number of contacts per electrode (default=1)")
@click.option(
    "--overhang",
    "-ov",
    default=None,
    type=float,
    help="extension (um) beyond MEA boundaries for neuron locations (default=30.)",
)
@click.option(
    "--xlim",
    "-xl",
    default=None,
    nargs=2,
    type=float,
    help="limits ( low high ) for neuron locations in the x-axis (depth) (default=[10.,80.])",
)
@click.option(
    "--ylim",
    "-yl",
    default=None,
    nargs=2,
    type=float,
    help="limits ( low high ) for neuron locations in the y-axis (default=None)",
)
@click.option(
    "--zlim",
    "-zl",
    default=None,
    nargs=2,
    type=float,
    help="limits ( low high ) for neuron locations in the z-axis (default=None)",
)
@click.option("--seed", "-s", default=None, type=int, help="random seed for template generation (int)")
@click.option("--intraonly", "-io", is_flag=True, help="only run intracellular simulations")
@click.option("--no-parallel", "-nopar", is_flag=True, help="run without multiprocessing tool")
@click.option("--recompile", "-rc", is_flag=True, help="recompile models")
@click.option("--drifting", "-dr", is_flag=True, help="generate drifting templates")
@click.option("--min-amp", "-mamp", type=float, help="miniumum template amplitude")
@click.option("--min-drift", "-mind", type=float, help="miniumum drifting distance")
@click.option("--max-drift", "-maxd", type=float, help="maximum drifting distance")
@click.option("--drift-steps", "-drst", type=int, help="number of drift steps")
@click.option(
    "--drift-xlim",
    "-dxl",
    default=None,
    nargs=2,
    type=float,
    help="limits ( low high ) for neuron drift locations in the x-axis (depth) (default=[10.,80.])",
)
@click.option(
    "--drift-ylim",
    "-dyl",
    default=None,
    nargs=2,
    type=float,
    help="limits ( low high ) for neuron drift locations in the y-axis (default=None)",
)
@click.option(
    "--drift-zlim",
    "-dzl",
    default=None,
    nargs=2,
    type=float,
    help="limits ( low high ) for neuron drift locations in the z-axis (default=None)",
)
@click.option("--verbose", "-v", is_flag=True, help="produce verbose output")
def gen_templates(params, **kwargs):
    """Generates TEMPLATES with biophysical simulation."""
    info, _ = get_default_config()

    if params is None:
        params = info["templates_params"]
    params_dict = safe_yaml_load(params)

    if kwargs["default"] is True:
        pprint.pprint(params_dict)
        mu.return_mea()
        return

    if kwargs["cellfolder"] is not None:
        model_folder = kwargs["cellfolder"]
    else:
        model_folder = info["cell_models_folder"]
    params_dict["cell_models_folder"] = model_folder

    if kwargs["seed"] is not None:
        seed = kwargs["seed"]
    else:
        seed = np.random.randint(1, 10000)
    params_dict["seed"] = seed

    if kwargs["folder"] is not None:
        templates_folder = kwargs["folder"]
        params_dict["templates_folder"] = Path(kwargs["folder"])
    else:
        templates_folder = Path(info["templates_folder"])
    intraonly = kwargs["intraonly"]

    if kwargs["rot"] is not None:
        params_dict["rot"] = kwargs["rot"]
    if kwargs["n"] is not None:
        params_dict["n"] = kwargs["n"]
    if kwargs["dt"] is not None:
        params_dict["dt"] = kwargs["dt"]
    if kwargs["ncontacts"] is not None:
        params_dict["ncontacts"] = kwargs["ncontacts"]
    if kwargs["overhang"] is not None:
        params_dict["overhang"] = kwargs["overhang"]
    if kwargs["xlim"] is not None and len(kwargs["xlim"]) == 2:
        params_dict["xlim"] = kwargs["xlim"]
    if kwargs["ylim"] is not None and len(kwargs["ylim"]) == 2:
        params_dict["ylim"] = kwargs["ylim"]
    if kwargs["zlim"] is not None and len(kwargs["zlim"]) == 2:
        params_dict["zlim"] = kwargs["zlim"]
    if kwargs["min_amp"] is not None:
        params_dict["min_amp"] = kwargs["min_amp"]

    if kwargs["drifting"]:
        params_dict["drifting"] = True
    elif "drifting" not in params_dict.keys():
        params_dict["drifting"] = False
    if kwargs["min_drift"] is not None:
        params_dict["min_drift"] = kwargs["min_drift"]
    if kwargs["max_drift"] is not None:
        params_dict["max_drift"] = kwargs["max_drift"]
    if kwargs["drift_steps"] is not None:
        params_dict["drift_steps"] = kwargs["drift_steps"]
    if kwargs["drift_xlim"] is not None and len(kwargs["drift_xlim"]) == 2:
        params_dict["drift_xlim"] = kwargs["drift_xlim"]
    if kwargs["drift_ylim"] is not None and len(kwargs["drift_ylim"]) == 2:
        params_dict["drift_ylim"] = kwargs["drift_ylim"]
    if kwargs["drift_zlim"] is not None and len(kwargs["drift_zlim"]) == 2:
        params_dict["drift_zlim"] = kwargs["drift_zlim"]

    if kwargs["probe"] is not None:
        if kwargs["probe"] in mu.return_mea_list():
            params_dict["probe"] = kwargs["probe"]
        else:
            print("The probe ", kwargs["probe"], " is not listead as a MEAutility probe.")
            return
    if kwargs["no_parallel"]:
        parallel = False
    else:
        parallel = True
    njobs = kwargs["njobs"]

    recompile = kwargs["recompile"]
    verbose = kwargs["verbose"]

    params_dict["templates_folder"] = str(templates_folder)

    tempgen = gt.gen_templates(
        cell_models_folder=model_folder,
        params=params_dict,
        templates_tmp_folder=str(templates_folder),
        intraonly=intraonly,
        n_jobs=njobs,
        recompile=recompile,
        parallel=parallel,
        verbose=verbose,
    )

    # Merge simulated data and cleanup
    if not intraonly:
        rot = params_dict["rot"]
        n = params_dict["n"]
        probe = params_dict["probe"]
        if kwargs["fname"] is None:
            if params_dict["drifting"]:
                fname = f'templates_{n}_{probe}_drift_{time.strftime("%d-%m-%Y_%H-%M")}.h5'
            else:
                fname = f'templates_{n}_{probe}_{time.strftime("%d-%m-%Y_%H-%M")}.h5'
        else:
            fname = kwargs["fname"]
        save_fname = str(templates_folder / rot / fname)
        save_template_generator(tempgen, save_fname, verbose=True)


@cli.command()
@click.option("--templates", "-t", default=None, help="templates path")
@click.option(
    "--params",
    "-prm",
    default=None,
    help="path to default_params.yaml (otherwise default default_params are used "
    "and some of the parameters can be overwritten with the following options)",
)
@click.option("--default", is_flag=True, help="shows default values for simulation")
@click.option("--fname", "-fn", default=None, help="recording filename")
@click.option("--folder", "-fol", default=None, help="recording output base folder")
@click.option("--duration", "-d", default=None, type=float, help="duration in s (default=10)")
@click.option("--njobs", "-nj", default=0, type=int, help="number of jobs (default n_jobs=n_cpus)")
@click.option("--n-exc", "-ne", default=None, type=int, help="number of excitatory cells (default=15)")
@click.option("--n-inh", "-ni", default=None, type=int, help="number of inhibitory cells (default=5)")
@click.option(
    "--f-exc", "-fe", default=None, type=float, help="average firing rate of excitatory cells in Hz (default=5)"
)
@click.option(
    "--f-inh", "-fi", default=None, type=float, help="average firing rate of inhibitory cells in Hz (default=15)"
)
@click.option(
    "--st-exc",
    "-se",
    default=None,
    type=float,
    help="firing rate standard deviation of excitatory cells in Hz (default=1)",
)
@click.option(
    "--st-inh",
    "-si",
    default=None,
    type=float,
    help="firing rate standard deviation of inhibitory cells in Hz (default=3)",
)
@click.option("--min-rate", "-mr", default=None, type=float, help="minimum firing rate (default=0.5)")
@click.option("--ref-per", "-rp", default=None, type=float, help="refractory period in ms (default=2)")
@click.option(
    "--process",
    "-p",
    default="poisson",
    type=click.Choice(["poisson", "gamma"]),
    help="process for generating spike trains (default=poisson)",
)
@click.option("--tstart", default=None, type=float, help="start time in s (default=0)")
@click.option("--min-dist", "-md", default=None, type=int, help="minumum distance between neuron in um (default=25)")
@click.option("--min-amp", "-mina", default=None, type=int, help="minumum eap amplitude in uV (default=50)")
@click.option("--max-amp", "-maxa", default=None, type=int, help="maximum eap amplitude in uV (default=inf)")
@click.option(
    "--fs", default=None, type=float, help="sampling frequency in kHz (default from templates sampling frequency)"
)
@click.option(
    "--sync-rate", "-sr", default=None, type=float, help="added synchrony rate on spatially overlapping spikes"
)
@click.option("--sync-jitt", "-sj", default=None, type=float, help="jitter in ms for added spikes")
@click.option("--noise-lev", "-nl", default=None, type=int, help="noise level in uV (default=10)")
@click.option(
    "--modulation", "-m", default=None, type=click.Choice(["none", "template", "electrode"]), help="modulation type"
)
@click.option("--chunk-duration", "-chd", default=None, type=float, help="chunk duration in s (default 0)")
@click.option("--st-seed", "-stseed", default=None, type=int, help="random seed for spike trains")
@click.option("--temp-seed", "-tseed", default=None, type=int, help="random seed for template selection")
@click.option("--conv-seed", "-cseed", default=None, type=int, help="random seed for convolution")
@click.option("--noise-seed", "-nseed", default=None, type=int, help="random seed for noise")
@click.option(
    "--half-dist",
    "-hd",
    default=None,
    type=float,
    help="when noise is distance-correlated the distance at which covariance is 0.5 (default=30)",
)
@click.option("--color-noise", "-cn", is_flag=True, help="if True noise is colored")
@click.option("--color-peak", "-cp", default=None, type=float, help="peak for noise in Hz (default 500 Hz)")
@click.option("--color-q", "-cq", default=None, type=float, help="quality factor for noise filter (default=1)")
@click.option(
    "--random-noise-floor", "-rnf", default=None, type=float, help="noise floor in std of additive noide (default 1)"
)
@click.option("--filter", is_flag=True, help="if True filter is applied")
@click.option(
    "--filt-cutoff",
    "-fc",
    default=None,
    type=float,
    multiple=True,
    help="filter cutoff frequencies." "High-pass: -fc hp-cutoff. Band-pass: -fc hp-cutoff -fc lp-cutoff",
)
@click.option("--filt-order", "-fo", default=None, type=int, help="filter order (default 3)")
@click.option("--overlap", is_flag=True, help="if True it annotates overlapping spikes")
@click.option("--overlap-thresh", "-ot", type=float, help="overlap threshold for spatial overlap")
@click.option("--extract-wf", is_flag=True, help="if True it annotates overlapping spikes")
@click.option("--bursting", is_flag=True, help="if True ISI-dependent modulation is performed")
@click.option("--shape-mod", is_flag=True, help="if True shape modulation is performed")
@click.option("--drifting", "-dr", is_flag=True, help="generate drifting recordings")
@click.option(
    "--preferred-dir",
    "-prd",
    default=None,
    nargs=3,
    type=float,
    help="preferred drifting direction (0 0 1 is positive z-direction)",
)
@click.option("--angle-tol", "-angt", type=float, help="angle tollerance for preferred direction")
@click.option("--drift-velocity", "-drvel", type=float, help="drift velocity in um/min")
@click.option("--t-start-drift", "-tsd", type=float, help="drifting start time in s")
@click.option("--verbose", "-v", is_flag=True, help="produce verbose output")
def gen_recordings(params, **kwargs):
    """Generates RECORDINGS from TEMPLATES."""
    # Retrieve default_params file
    info, _ = get_default_config()

    if params is None:
        params = info["recordings_params"]
    params_dict = safe_yaml_load(params)

    if kwargs["default"] is True:
        pprint.pprint(params_dict)
        return

    templates_folder = Path(kwargs["templates"])

    if kwargs["folder"] is not None:
        params_dict["recordings"].update({"recordings_folder": kwargs["folder"]})
    else:
        params_dict["recordings"].update({"recordings_folder": info["recordings_folder"]})
    recordings_folder = params_dict["recordings"]["recordings_folder"]

    if kwargs["templates"] is None:
        print("Provide eap templates path")
        return
    else:
        if templates_folder.is_dir() or templates_folder.suffix in [".h5", ".hdf5"]:
            params_dict["templates"].update({"templates": kwargs["templates"]})
        else:
            print(kwargs["templates"])
            raise AttributeError("'templates' is not a folder")

    if kwargs["n_exc"] is not None:
        params_dict["spiketrains"]["n_exc"] = kwargs["n_exc"]
    if kwargs["n_inh"] is not None:
        params_dict["spiketrains"]["n_inh"] = kwargs["n_inh"]
    if kwargs["f_exc"] is not None:
        params_dict["spiketrains"]["f_exc"] = kwargs["f_exc"]
    if kwargs["f_inh"] is not None:
        params_dict["spiketrains"]["f_inh"] = kwargs["f_inh"]
    if kwargs["st_exc"] is not None:
        params_dict["spiketrains"]["st_exc"] = kwargs["st_exc"]
    if kwargs["st_inh"] is not None:
        params_dict["spiketrains"]["st_inh"] = kwargs["st_inh"]
    if kwargs["min_rate"] is not None:
        params_dict["spiketrains"]["min_rate"] = kwargs["min_rate"]
    if kwargs["ref_per"] is not None:
        params_dict["spiketrains"]["ref_per"] = kwargs["ref_per"]
    if kwargs["process"] is not None:
        params_dict["spiketrains"]["process"] = kwargs["process"]
    if kwargs["min_rate"] is not None:
        params_dict["spiketrains"]["min_rate"] = kwargs["min_rate"]
    if kwargs["duration"] is not None:
        params_dict["spiketrains"]["duration"] = kwargs["duration"]
    if kwargs["tstart"] is not None:
        params_dict["spiketrains"]["t_start"] = kwargs["t_start"]
    if kwargs["min_dist"] is not None:
        params_dict["templates"]["min_dist"] = kwargs["min_dist"]
    if kwargs["min_amp"] is not None:
        params_dict["templates"]["min_amp"] = kwargs["min_amp"]
    if kwargs["max_amp"] is not None:
        params_dict["templates"]["max_amp"] = kwargs["max_amp"]
    if kwargs["overlap_thresh"] is not None:
        params_dict["templates"]["overlap_threshold"] = kwargs["overlap_thresh"]

    if kwargs["noise_lev"] is not None:
        params_dict["recordings"]["noise_level"] = kwargs["noise_lev"]
    if kwargs["modulation"] is not None:
        params_dict["recordings"]["modulation"] = kwargs["modulation"]

    if kwargs["chunk_duration"] is not None:
        params_dict["recordings"]["chunk_duration"] = kwargs["chunk_duration"]
    if kwargs["filter"]:
        params_dict["recordings"]["filter"] = True
    else:
        params_dict["recordings"]["filter"] = False
    if kwargs["filt_cutoff"] is not None:
        if isinstance(kwargs["filt_cutoff"], tuple):
            kwargs["filt_cutoff"] = list(kwargs["filt_cutoff"])
        # for high-pass filter, make a scalar
        if len(kwargs["filt_cutoff"]) == 1:
            kwargs["filt_cutoff"] = kwargs["filt_cutoff"][0]
        params_dict["recordings"]["filter_cutoff"] = kwargs["filt_cutoff"]
    if kwargs["filt_order"] is not None:
        params_dict["recordings"]["filt_order"] = kwargs["filt_order"]
    if kwargs["fs"] is not None:
        params_dict["recordings"]["fs"] = kwargs["fs"]
    else:
        params_dict["recordings"]["fs"] = None
    if kwargs["sync_rate"] is not None:
        params_dict["recordings"]["sync_rate"] = kwargs["sync_rate"]
    else:
        params_dict["recordings"]["sync_rate"] = None
    if kwargs["sync_jitt"] is not None:
        params_dict["recordings"]["sync_jitt"] = kwargs["sync_jitt"]
    else:
        params_dict["recordings"]["sync_jitt"] = 1
    if kwargs["noise_seed"] is not None:
        params_dict["seeds"]["noise"] = kwargs["noise_seed"]
    if kwargs["temp_seed"] is not None:
        params_dict["seeds"]["templates"] = kwargs["temp_seed"]
    if kwargs["st_seed"] is not None:
        params_dict["seeds"]["spiketrains"] = kwargs["st_seed"]
    if kwargs["conv_seed"] is not None:
        params_dict["seeds"]["convolution"] = kwargs["conv_seed"]
    if kwargs["overlap"]:
        params_dict["recordings"]["overlap"] = True
    elif "overlap" not in params_dict["recordings"].keys():
        params_dict["recordings"]["overlap"] = False
    if kwargs["extract_wf"]:
        params_dict["recordings"]["extract_wf"] = True
    elif "extract_wf" not in params_dict["recordings"].keys():
        params_dict["recordings"]["extract_wf"] = False

    if kwargs["half_dist"] is not None:
        params_dict["recordings"]["half_dist"] = kwargs["half_dist"]
    if kwargs["color_noise"]:
        params_dict["recordings"]["noise_color"] = True
    elif "noise_color" not in params_dict["recordings"].keys():
        params_dict["recordings"]["noise_color"] = False
    if kwargs["color_peak"] is not None:
        params_dict["recordings"]["color_peak"] = kwargs["color_peak"]
    if kwargs["color_q"] is not None:
        params_dict["recordings"]["color_q"] = kwargs["color_q"]
    if kwargs["random_noise_floor"] is not None:
        params_dict["recordings"]["random_noise_floor"] = kwargs["random_noise_floor"]

    if kwargs["bursting"]:
        params_dict["recordings"]["bursting"] = True
    elif "bursting" not in params_dict["recordings"].keys():
        params_dict["recordings"]["bursting"] = False

    if kwargs["shape_mod"]:
        params_dict["recordings"]["shape_mod"] = True
    elif "shape_mod" not in params_dict["recordings"].keys():
        params_dict["recordings"]["shape_mod"] = False

    if kwargs["drifting"]:
        params_dict["recordings"]["drifting"] = True
    elif "drifting" not in params_dict["recordings"].keys():
        params_dict["recordings"]["drifting"] = False
    if kwargs["preferred_dir"] is not None and len(kwargs["preferred_dir"]) == 3:
        params_dict["recordings"]["preferred_dir"] = kwargs["preferred_dir"]
    if kwargs["angle_tol"]:
        params_dict["recordings"]["angle_tol"] = kwargs["angle_tol"]
    if kwargs["drift_velocity"]:
        params_dict["recordings"]["drift_velocity"] = kwargs["drift_velocity"]
    if kwargs["t_start_drift"]:
        params_dict["recordings"]["t_start_drift"] = kwargs["t_start_drift"]
    if kwargs["verbose"]:
        verbose = 2
    else:
        verbose = kwargs["verbose"]
    njobs = kwargs["njobs"]

    recgen = gt.gen_recordings(templates=kwargs["templates"], params=params_dict, verbose=verbose, n_jobs=njobs)
    info = recgen.info

    n_neurons = info["recordings"]["n_neurons"]
    electrode_name = info["electrodes"]["electrode_name"]
    duration = info["recordings"]["duration"]
    noise_level = info["recordings"]["noise_level"]

    if kwargs["fname"] is None:
        if kwargs["drifting"]:
            fname = (
                f"recordings_{n_neurons}cells_{electrode_name}_{duration}_{np.round(noise_level, 2)}uV_"
                f'drift_{time.strftime("%d-%m-%Y_%H-%M")}.h5'
            )
        else:
            fname = (
                f"recordings_{n_neurons}cells_{electrode_name}_{duration}_{np.round(noise_level, 2)}uV_"
                f'{time.strftime("%d-%m-%Y_%H-%M")}.h5'
            )
    else:
        fname = kwargs["fname"]

    if recordings_folder is not None:
        recordings_folder = Path(recordings_folder)
        if not recordings_folder.is_dir():
            os.makedirs(str(recordings_folder))
    rec_path = str(recordings_folder / fname)
    save_recording_generator(recgen, rec_path, verbose=True)


@cli.command()
def default_config():
    """Print default configurations."""
    info, _ = get_default_config(print_version=True)
    pprint.pprint(info)


@cli.command()
@click.option("--info", is_flag=True, default=False, help="if True probe information is printed")
def available_probes(info):
    """Print available probes."""
    from pprint import pprint

    probe_list = mu.return_mea_list()
    for p in probe_list:
        probe_info = mu.return_mea_info(p)
        if "description" not in probe_info.keys():
            probe_info["description"] = "no description"
        if info:
            print(p)
            pprint(mu.return_mea_info(p))
        else:
            print(p, "----------", probe_info["description"])
        print()


@cli.command()
@click.argument("cell-models-folder")
def set_cell_models_folder(cell_models_folder):
    """Set default cell_models folder."""
    info, config = get_default_config()
    cell_models_folder = Path(cell_models_folder)
    if cell_models_folder.is_dir():
        info["cell_models_folder"] = str(cell_models_folder.absolute())
        with open(Path(config) / "mearec.conf", "w") as f:
            yaml.dump(info, f)
        print("Set default cell_models_folder to: ", cell_models_folder)
    else:
        print(cell_models_folder, " is not a folder!")


@cli.command()
@click.argument("templates-folder")
@click.option("--create", is_flag=True, help="if True it creates the directory")
def set_templates_folder(templates_folder, create):
    """Set default templates output folder."""
    info, config = get_default_config()
    templates_folder = Path(templates_folder).absolute()
    if templates_folder.is_dir():
        info["templates_folder"] = str(templates_folder)
        with open(Path(config) / "mearec.conf", "w") as f:
            yaml.dump(info, f)
        print("Set default templates_folder to: ", templates_folder)
    elif create:
        os.makedirs(str(templates_folder))
        info["templates_folder"] = str(templates_folder)
        with open(Path(config) / "mearec.conf", "w") as f:
            yaml.dump(info, f)
        print("Set default templates_folder to: ", templates_folder)
    else:
        print(templates_folder, " is not a folder!")


@cli.command()
@click.argument("recordings-folder")
@click.option("--create", is_flag=True, help="if True it creates the directory")
def set_recordings_folder(recordings_folder, create):
    """Set default recordings output folder."""
    info, config = get_default_config()
    recordings_folder = Path(recordings_folder).absolute()
    if recordings_folder.is_dir():
        info["recordings_folder"] = str(recordings_folder)
        with open(Path(config) / "mearec.conf", "w") as f:
            yaml.dump(info, f)
        print("Set default recordings_folder to: ", recordings_folder)
    elif create:
        os.makedirs(str(recordings_folder))
        info["recordings_folder"] = str(recordings_folder)
        with open(Path(config) / "mearec.conf", "w") as f:
            yaml.dump(info, f)
        print("Set default recordings_folder to: ", recordings_folder)
    else:
        print(recordings_folder, " is not a folder!")


@cli.command()
@click.argument("templates-params")
def set_templates_params(templates_params):
    """Set default templates parameter file."""
    info, config = get_default_config()
    templates_params = Path(templates_params).absolute()
    if templates_params.is_file() and templates_params.suffix in [".yaml", ".yml"]:
        info["templates_params"] = str(templates_params)
        with open(Path(config) / "mearec.conf", "w") as f:
            yaml.dump(info, f)
        print("Set default templates_params to: ", templates_params)
    else:
        print(templates_params, " is not a yaml!")


@cli.command()
@click.argument("recordings-params")
def set_recordings_params(recordings_params):
    """Set default recordings parameter file."""
    info, config = get_default_config()
    recordings_params = Path(recordings_params).absolute()
    if recordings_params.is_file() and recordings_params.suffix in [".yaml", ".yml"]:
        info["recording_params"] = str(recordings_params)
        with open(Path(config) / "mearec.conf", "w") as f:
            yaml.dump(info, f)
        print("Set default recordings_params to: ", recordings_params)
    else:
        print(recordings_params, " is not a yaml!")



"""
Test implementation using cell models of the Blue Brain Project with LFPy.
The example assumes that cell models available from
https://bbpnmc.epfl.ch/nmc-portal/downloads are unzipped in the folder 'cell_models'

The function compile_all_mechanisms must be run once before any cell simulation
"""

import os
import sys
import time
from pathlib import Path

import MEAutility as mu
import numpy as np
from packaging.version import parse

from MEArec.tools import safe_yaml_load


def import_LFPy_neuron():
    try:
        import LFPy
    except:
        raise ModuleNotFoundError("LFPy is not installed. Install it with 'pip install LFPy'")

    try:
        # disable DISPLAY for subprocess
        # print("Disabling display")
        # os.environ["DISPLAY"] = ""
        import neuron
    except:
        raise ModuleNotFoundError(
            "NEURON is not installed. Install it from https://www.neuron.yale.edu/neuron/download"
        )

    if parse(LFPy.__version__) < parse("2.2"):
        raise ImportError("LFPy version must be >= 2.2. To use a previous LFPy version, downgrade MEArec to <= 1.4.1")

    return LFPy, neuron


def get_templatename(f):
    """
    Assess from hoc file the templatename being specified within

    Arguments
    ---------
    f : file, mode 'r'

    Returns
    -------
    templatename : str

    """
    templatename = None
    f = open("template.hoc", "r")
    for line in f.readlines():
        if "begintemplate" in line.split():
            templatename = line.split()[-1]
            continue
    return templatename


def compile_all_mechanisms(cell_folder, verbose=False):
    """Attempt to set up a folder with all unique mechanism *.mod files and
        compile them all. assumes all cell models are in a folder 'cell_models'

    Parameters
    -----------
    cell_folder : str
        Path to cell folder
    """
    cell_folder = Path(cell_folder)
    mod_folder = cell_folder / "mods"
    mod_folder.mkdir(exist_ok=True, parents=True)

    neurons = [f for f in cell_folder.iterdir() if "mods" not in str(f) and not f.name.startswith(".")]

    if verbose >= 1:
        print(neurons)

    for neuron in neurons:
        for nmodl in (neuron / "mechanisms").iterdir():
            if nmodl.suffix == ".mod":
                while not (cell_folder / "mods" / nmodl.parts[-1]).is_file():
                    if sys.platform == "win32":
                        _command = "copy"
                    else:
                        _command = "cp"
                    if verbose >= 1:
                        print(f"{_command} {nmodl} {cell_folder / 'mods'}")
                    os.system(f"{_command} {nmodl} {cell_folder / 'mods'}")
    starting_dir = os.getcwd()
    os.chdir(str(cell_folder / "mods"))
    os.system("nrnivmodl")
    os.chdir(starting_dir)


def return_bbp_cell(cell_folder, end_T, dt, start_T, verbose=0):
    """Function to load cell models

    Parameters
    ----------
    cell_folder : string
        Path to folder with the BBP cell model
    end_T : float
        Simulation length [ms]
    dt: float
        Time step of simulation [ms]
    start_T: float
        Simulation start time (recording starts at 0 ms)

    Returns
    -------
    cell : object
        LFPy cell object
    """
    LFPy, neuron = import_LFPy_neuron()
    neuron.h.load_file("stdrun.hoc")
    neuron.h.load_file("import3d.hoc")

    cwd = os.getcwd()
    os.chdir(cell_folder)
    if verbose >= 1:
        print(f"Simulating {cell_folder}")

    neuron.load_mechanisms(str(Path(cell_folder).parent / "mods"))

    f = open("template.hoc", "r")
    templatename = get_templatename(f)
    f.close()

    f = open("biophysics.hoc", "r")
    biophysics = get_templatename(f)
    f.close()

    f = open("morphology.hoc", "r")
    morphology = get_templatename(f)
    f.close()

    # get synapses template name
    synapses_file = str(Path("synapses") / "synapses.hoc")
    f = open(synapses_file, "r")
    synapses = get_templatename(f)
    f.close()

    neuron.h.load_file("constants.hoc")
    if not hasattr(neuron.h, morphology):
        neuron.h.load_file(1, "morphology.hoc")

    if not hasattr(neuron.h, biophysics):
        neuron.h.load_file(1, "biophysics.hoc")

    if not hasattr(neuron.h, synapses):
        # load synapses
        neuron.h.load_file(1, synapses_file)

    if not hasattr(neuron.h, templatename):
        neuron.h.load_file(1, "template.hoc")

    morphologyfile = [f for f in Path("morphology").iterdir()][0]

    # Instantiate the cell(s) using LFPy
    cell = LFPy.TemplateCell(
        morphology=str(morphologyfile),
        templatefile=str(Path("template.hoc").absolute()),
        templatename=templatename,
        templateargs=0,
        tstop=end_T,
        tstart=start_T,
        dt=dt,
        v_init=-70,
        pt3d=True,
        delete_sections=True,
        verbose=True,
    )

    os.chdir(cwd)
    return cell


def return_bbp_cell_morphology(cell_name, cell_folder, pt3d=False):
    """Function to load cell models

    Parameters
    ----------
    cell_name : string
        Name of the cell type.
    cell_folder : string
        Folder containing cell models.
    pt3d : bool
        If True detailed 3d morphology is used

    Returns
    -------
    cell : object
        LFPy cell object
    """
    LFPy, neuron = import_LFPy_neuron()
    cell_folder = Path(cell_folder)

    if not (cell_folder / cell_name).is_dir():
        raise NotImplementedError(f"Cell model {cell_name} is not found in {cell_folder}")

    morphology_files = [f for f in (cell_folder / cell_name / "morphology").iterdir()]
    if len(morphology_files) > 1:
        raise Exception(f"More than 1 morphology file found for cell {cell_name}")
    morphology = morphology_files[0]

    cell = LFPy.Cell(morphology=str(morphology), pt3d=pt3d)
    return cell


def find_spike_idxs(v, thresh=-30, find_max=30):
    """Find spike indices

    Parameters
    ----------
    v: array_like
        Membrane potential
    thresh: float (optional, default = -30)
        Threshold for spike detections
    find_max: int
        Number of sample to find spike maximum after detection crossing

    Returns
    -------
    spikes : array_like
        Indices of spike peaks in the positive direction, i.e. spikes
    """
    spikes_th = [idx for idx in range(len(v) - 1) if v[idx] < thresh < v[idx + 1]]
    spikes = []
    for sp in spikes_th:
        max_idx = np.argmax(v[sp : sp + find_max])
        spikes.append(sp + max_idx)
    return spikes


def set_input(weight, dt, T, cell, delay, stim_length):
    """Set current input synapse in soma

    Parameters
    ----------
    weight : float
        Strength of input current [nA]
    dt : float
        Time step of simulation [ms]
    T : float
        Total simulation time [ms]
    cell : object
        Cell object from LFPy
    delay : float
        Delay for input,i.e. when to start the input [ms]
    stim_length: float
        Duration of injected current [ms]

    Returns
    -------
    noiseVec :  NEURON vector
        NEURON vector of input current
    cell : object
        LFPy cell object
    synapse : NEURON synapse
        NEURON synapse object
    """
    import neuron

    tot_ntsteps = int(round(T / dt + 1))

    I = np.ones(tot_ntsteps) * weight
    # I[stim_idxs] = weight
    noiseVec = neuron.h.Vector(I)
    syn = None
    for sec in cell.allseclist:
        if "soma" in sec.name():
            # syn = neuron.h.ISyn(0.5, sec=sec)
            syn = neuron.h.IClamp(0.5, sec=sec)
    syn.dur = stim_length
    syn.delay = delay  # cell.tstartms
    noiseVec.play(syn._ref_amp, dt)

    return noiseVec, cell, syn


def run_cell_model(
    cell_model_folder, verbose=False, sim_folder=None, save=True, custom_return_cell_function=None, **kwargs
):
    """
    Run simulation and adjust input strength to have a certain number of
    spikes (target_spikes[0] < num_spikes <= target_spikes[1]
    where target_spikes=[10,30] by default)

    Parameters
    ----------
    cell_model_folder : string
        Path to folder where cell model is saved.
    verbose : int
        If 1, output is verbose
    save : bool
        If True, currents and membrane potentials are saved in 'sim_folder'. If False the function returns the simulated
        cell, the soma potentials of the spikes, and the transmembrane currents of the spikes
    sim_folder : string
        Data directory for transmembrane currents and membrane potential of the neuron.
    custom_return_cell_function : function
        Python function to to return an LFPy cell from the cell_model_folder
    **kwargs : keyword arguments
        Kwargs must include: 'sim_time', 'dt', 'delay', 'weights', 'target_spikes', 'cut_out', 'seed'

    Returns
    -------
    cell : object
        LFPy cell object (if save is False)
    v : np.array
        Array (N_spikes x t) with soma membrane potential (if save is False)
    i : np.array
        Array (N_spikes x N_compartments x t) with transmembrane currents (if save is False)

    """
    cell_name = Path(cell_model_folder).parts[-1]
    if sim_folder is not None:
        sim_folder = Path(sim_folder)

    if custom_return_cell_function is None:
        return_function = return_bbp_cell
    else:
        return_function = custom_return_cell_function

    intra_params = kwargs

    if save:
        assert sim_folder is not None, "Specify 'save_sim_folder' argument!"
        sim_folder.mkdir(exist_ok=True, parents=True)

        imem_names = [f.name for f in sim_folder.iterdir() if "imem" in f.name]
        vmem_names = [f.name for f in sim_folder.iterdir() if "vmem" in f.name]
    else:
        imem_names = []
        vmem_names = []

    if not (
        np.any([cell_name in iname for iname in imem_names]) and np.any([cell_name in vname for vname in vmem_names])
    ):
        np.random.seed(intra_params["seed"])
        T = intra_params["sim_time"] * 1000
        dt = intra_params["dt"]
        cell = return_function(cell_model_folder, end_T=T, dt=dt, start_T=0)

        delay = intra_params["delay"]
        stim_length = T - delay
        weights = intra_params["weights"]
        weight = weights[0]
        target_spikes = intra_params["target_spikes"]
        cuts = intra_params["cut_out"]
        cut_out = [int(cuts[0] / dt), int(cuts[1] / dt)]

        num_spikes = 0

        i = 0
        while not target_spikes[0] < num_spikes <= target_spikes[1]:
            noiseVec, cell, syn = set_input(weight, dt, T, cell, delay, stim_length)
            cell.simulate(rec_imem=True)

            t = cell.tvec
            v = cell.somav
            t = t
            v = v

            spikes = find_spike_idxs(v[cut_out[0] : -cut_out[1]])
            spikes = list(np.array(spikes) + cut_out[0])
            num_spikes = len(spikes)

            if verbose >= 1:
                print(f"Input weight: {weight} - Num Spikes: {num_spikes}")
            if num_spikes >= target_spikes[1]:
                weight *= weights[0]
            elif num_spikes <= target_spikes[0]:
                weight *= weights[1]

            i += 1
            if i >= 10:
                sys.exit()

        t = t[0 : (cut_out[0] + cut_out[1])] - t[cut_out[0]]
        # discard first spike
        i_spikes = np.zeros((num_spikes - 1, cell.totnsegs, len(t)))
        v_spikes = np.zeros((num_spikes - 1, len(t)))

        for idx, spike_idx in enumerate(spikes[1:]):
            spike_idx = int(spike_idx)
            v_spike = v[spike_idx - cut_out[0] : spike_idx + cut_out[1]]
            i_spike = cell.imem[:, spike_idx - cut_out[0] : spike_idx + cut_out[1]]
            i_spikes[idx, :, :] = i_spike
            v_spikes[idx, :] = v_spike
        if save:
            np.save(str(sim_folder / f"imem_{num_spikes - 1}_{cell_name}.npy"), i_spikes)
            np.save(str(sim_folder / f"vmem_{num_spikes - 1}_{cell_name}.npy"), v_spikes)
        else:
            return cell, v_spikes, i_spikes
    else:
        if verbose >= 1:
            print("\n\n\nCell has already be simulated. Using stored membrane currents\n\n\n")


def calculate_extracellular_potential(cell, mea, ncontacts=10, position=None, rotation=None):
    """
    Calculates extracellular signal in uV on MEA object.

    Parameters
    ----------
    cell : LFPy Cell
        The simulated cell
    mea : MEA, str, or dict
        Mea object from MEAutility, string with probe name, or dict with probe info

    Returns
    -------
    v_ext : np.array
        Extracellular potential computed on the electrodes (n_elec, n_timestamps)
    """
    LFPy, neuron = import_LFPy_neuron()

    if isinstance(mea, str):
        mea_obj = mu.return_mea(mea)
    elif isinstance(mea, dict):
        mea_obj = mu.return_mea(info=mea)
    elif isinstance(mea, mu.core.MEA):
        mea_obj = mea
    else:
        raise Exception("")

    if ncontacts > 1:
        electrodes = LFPy.RecExtElectrode(cell, probe=mea_obj, n=ncontacts)
    else:
        electrodes = LFPy.RecExtElectrode(cell, probe=mea_obj)

    if position is not None:
        assert len(position) == 3, "'position' should be a 3d array"
        cell.set_pos(position[0], position[1], position[2])

    if rotation is not None:
        assert len(rotation) == 3, "'rotation' should be a 3d array"
        cell.set_rotation(x=rotation[0], y=rotation[1], z=rotation[2])

    lfp = electrodes.get_transformation_matrix() @ cell.imem

    # Reverse rotation to bring cell back into initial rotation state
    if rotation is not None:
        rev_rot = [-r for r in rotation]
        cell.set_rotation(rev_rot[0], rev_rot[1], rev_rot[2], rotation_order="zyx")

    return 1000 * lfp


def calc_extracellular(
    i,
    cell_model_folder,
    load_sim_folder,
    save_sim_folder=None,
    seed=0,
    verbose=0,
    position=None,
    custom_return_cell_function=None,
    cell_locations=None,
    cell_rotations=None,
    save=True,
    max_iterations=1000,
    timeout=300,
    **kwargs,
):
    """
    Loads data from previous cell simulation, and use results to generate
    arbitrary number of spikes above a certain noise level.

    Parameters
    ----------
    i: int
        Index of cell model
    cell_model_folder : string
        Path to folder where cell model is saved.
    model_type : string
        Cell model type (e.g. 'bbp')
    load_sim_folder : string
        Path to folder from which NEURON simulation results (currents, membrane potential) are loaded
    save_sim_folder : string
        Path to folder where to save EAP data
    position : array
        3D position of the soma (optional, default is None and the cell is randomly located within specified limits)
    custom_return_cell_function : function
        Python function to to return an LFPy cell from the cell_model_folder
    save : bool
        If True eaps are saved in the 'save_sim_folder'. If False eaps, positions, and rotations are returned as arrays
    cell_locations: np.array or None
        If passed, the passed locations are used (instead of random generation)
    cell_rotations: np.array or None
        If passed, the passed rotations are used (instead of random generation). Must be set if cell_locations is not
        None
    timeout : int
        Set the timeout in seconds for finding spikes over min_amp. Default 300
    max_iterations : int
        Set the maximum number of iterations for finding spikes over min_amp. Default 1000
    **kwargs: keyword arguments
        Template generation parameters (use mr.get_default_template_parameters() to retrieve the arguments)

    """
    LFPy, neuron = import_LFPy_neuron()

    for sec in neuron.h.allsec():
        neuron.h("%s{delete_section()}" % sec.name())

    cell_model_folder = Path(cell_model_folder)
    cell_name = cell_model_folder.parts[-1]
    cell_save_name = cell_name
    load_sim_folder = Path(load_sim_folder)
    if verbose >= 1:
        print(f"Seed = {seed + i}")
    np.random.seed(seed + i)

    T = kwargs["sim_time"] * 1000
    dt = kwargs["dt"]
    rotation = kwargs["rot"]
    nobs = kwargs["n"]
    ncontacts = kwargs["ncontacts"]
    overhang = kwargs["overhang"]
    x_lim = kwargs["xlim"]
    y_lim = kwargs["ylim"]
    z_lim = kwargs["zlim"]
    x_distr = kwargs["x_distr"]
    beta_distr_params = kwargs["beta_distr_params"]
    min_amp = kwargs["min_amp"]
    check_shape = kwargs["check_eap_shape"]
    MEAname = kwargs["probe"]
    drifting = kwargs["drifting"]

    if drifting:
        max_drift = kwargs["max_drift"]
        min_drift = kwargs["min_drift"]
        drift_steps = kwargs["drift_steps"]
        drift_x_lim = kwargs["drift_xlim"]
        drift_y_lim = kwargs["drift_ylim"]
        drift_z_lim = kwargs["drift_zlim"]
        check_for_drift_amp = kwargs["check_for_drift_amp"]
        drift_within_bounds = kwargs["drift_within_bounds"]

    if "timeout" in kwargs:
        timeout = kwargs.get("timeout")
    if "max_iterations" in kwargs:
        max_iterations = kwargs.get("max_iterations")

    if custom_return_cell_function is None:
        return_function = return_bbp_cell
        model_type = "bbp"
    else:
        return_function = custom_return_cell_function
        model_type = "custom"

    cuts = kwargs["cut_out"]
    cut_out = [int(cuts[0] / dt), int(cuts[1] / dt)]

    cell = return_function(cell_model_folder, end_T=T, dt=dt, start_T=0)

    # Load data from previous cell simulation
    imem_file = [f for f in load_sim_folder.iterdir() if cell_name in f.name and "imem" in f.name][0]
    vmem_file = [f for f in load_sim_folder.iterdir() if cell_name in f.name and "vmem" in f.name][0]
    i_spikes = np.load(str(imem_file))
    v_spikes = np.load(str(vmem_file))
    cell.tvec = np.arange(i_spikes.shape[-1]) * dt

    saved_eaps = []
    saved_positions = []
    saved_rotations = []
    target_num_spikes = int(nobs)

    # load MEA info
    elinfo = mu.return_mea_info(electrode_name=MEAname)

    # Create save folder
    if save:
        assert save_sim_folder is not None, "Specify 'save_sim_folder' argument!"
        save_sim_folder = Path(save_sim_folder)
        sim_folder = save_sim_folder / rotation
        save_folder = sim_folder / f"tmp_{target_num_spikes}_{MEAname}"
        save_folder.mkdir(exist_ok=True, parents=True)

    if verbose >= 1:
        print(f"Cell {cell_save_name} extracellular spikes to be simulated")

    mea = mu.return_mea(MEAname)
    if ncontacts > 1:
        electrodes = LFPy.RecExtElectrode(cell, probe=mea, n=ncontacts)
    else:
        electrodes = LFPy.RecExtElectrode(cell, probe=mea)
    pos = mea.positions
    elec_x = pos[:, 0]
    elec_y = pos[:, 1]
    elec_z = pos[:, 2]

    if x_lim is None:
        x_lim = [float(np.min(elec_x) - overhang), float(np.max(elec_x) + overhang)]
    if y_lim is None:
        y_lim = [float(np.min(elec_y) - overhang), float(np.max(elec_y) + overhang)]
    if z_lim is None:
        z_lim = [float(np.min(elec_z) - overhang), float(np.max(elec_z) + overhang)]

    # pre-simulate positions
    n_rand_positions = int(target_num_spikes * 1e5)
    print(f"Pre-generating {n_rand_positions} random positions")

    if x_distr == "uniform":
        x_rands = np.random.uniform(x_lim[0], x_lim[1], n_rand_positions)
    else:
        x_rands = x_lim[0] + np.random.beta(beta_distr_params[0], beta_distr_params[1], n_rand_positions) * (
            x_lim[1] - x_lim[0]
        )
    y_rands = np.random.uniform(y_lim[0], y_lim[1], n_rand_positions)
    z_rands = np.random.uniform(z_lim[0], z_lim[1], n_rand_positions)

    saved = 0
    i = 0
    tested_loc_idx = 0
    saved_amplitudes = []

    if cell_locations is not None:
        assert cell_rotations is not None, "If 'cell_locations' is not None, 'cell_rotations' should be given"

    if cell_locations is None:
        start_time = time.time()
        while len(saved_eaps) < target_num_spikes and tested_loc_idx < n_rand_positions:
            if i > max_iterations * target_num_spikes:
                if verbose >= 1:
                    print(f"Gave up finding spikes above noise level for {cell_name}")
                break
            # Each cell has several spikes to choose from
            spike_idx = np.random.randint(0, i_spikes.shape[0])
            cell.imem = i_spikes[spike_idx, :, :]
            cell.somav = v_spikes[spike_idx, :]
            tested_pos = [x_rands[tested_loc_idx], y_rands[tested_loc_idx], z_rands[tested_loc_idx]]

            espikes, pos, rot, found_position = return_extracellular_spike(
                cell=cell,
                cell_name=cell_name,
                model_type=model_type,
                electrodes=electrodes,
                limits=[x_lim, y_lim, z_lim],
                rotation=rotation,
                saved_pos=saved_positions,
                pos=tested_pos,
                verbose=False,
            )
            tested_loc_idx += 1
            if not found_position:
                continue

            # Method of Images for semi-infinite planes
            if elinfo["type"] == "mea":
                espikes = espikes * 2

            if not drifting:
                if check_espike(espikes, min_amp, check_shape):
                    skip = skip_duplicate(pos, saved_positions, drifting, verbose)
                    if skip:
                        continue

                    espikes = center_espike(espikes, cut_out)
                    saved_eaps.append(espikes)
                    saved_positions.append(pos)
                    saved_rotations.append(rot)
                    if verbose >= 1:
                        print(f"Cell: {cell_name} Progress: [{len(saved_eaps)}/{target_num_spikes}]")
                    saved += 1
            else:
                if check_espike(espikes, min_amp, check_shape):
                    if verbose >= 2:
                        print("template amplitude:", np.round(np.abs(np.min(espikes)), 1))

                    drift_ok = False
                    # fix rotation while drifting
                    cell.set_rotation(rot[0], rot[1], rot[2])
                    max_trials = 100
                    tr = 0

                    while not drift_ok and tr < max_trials:
                        init_pos = pos
                        # find drifting final position within drift limits
                        x_rand = np.random.uniform(init_pos[0] + drift_x_lim[0], init_pos[0] + drift_x_lim[1])
                        y_rand = np.random.uniform(init_pos[1] + drift_y_lim[0], init_pos[1] + drift_y_lim[1])
                        z_rand = np.random.uniform(init_pos[2] + drift_z_lim[0], init_pos[2] + drift_z_lim[1])
                        final_pos = [x_rand, y_rand, z_rand]
                        drift_dist = np.linalg.norm(np.array(init_pos) - np.array(final_pos))

                        # check location and boundaries
                        if drift_within_bounds:
                            if not (
                                x_lim[0] < x_rand < x_lim[1]
                                and y_lim[0] < y_rand < y_lim[1]
                                and z_lim[0] < z_rand < z_lim[1]
                            ):
                                if verbose == 2:
                                    print(f"Discarded for final drift position {cell_name}")
                                tr += 1
                                continue
                        if max_drift >= drift_dist >= min_drift:
                            if check_for_drift_amp:
                                # check final position spike amplitude
                                espikes, pos, rot_, found_position = return_extracellular_spike(
                                    cell=cell,
                                    cell_name=cell_name,
                                    model_type=model_type,
                                    electrodes=electrodes,
                                    limits=[x_lim, y_lim, z_lim],
                                    rotation=None,
                                    saved_pos=saved_positions,
                                    pos=final_pos,
                                )
                                if not found_position:
                                    continue

                                # Method of Images for semi-infinite planes
                                if elinfo["type"] == "mea":
                                    espikes = espikes * 2

                                if check_espike(espikes, min_amp, check_shape):
                                    if verbose == 2:
                                        print("Found final drifting position")
                                    drift_ok = True
                                else:
                                    tr += 1
                                    if verbose == 2:
                                        print(f"Discarded for final drift amplitude {cell_name}")
                                    continue
                            else:
                                drift_ok = True
                        else:
                            tr += 1
                            if verbose == 2:
                                print(f"Discarded for drift distance {cell_name}")

                    # now compute drifting templates
                    if drift_ok:
                        drift_spikes = []
                        drift_pos = []
                        drift_dir = np.array(final_pos) - np.array(init_pos)
                        for i, dp in enumerate(np.linspace(0, 1, drift_steps)):
                            pos_drift = init_pos + dp * drift_dir
                            espikes, pos, r_, found_position = return_extracellular_spike(
                                cell=cell,
                                cell_name=cell_name,
                                model_type=model_type,
                                electrodes=electrodes,
                                limits=[x_lim, y_lim, z_lim],
                                rotation=None,
                                saved_pos=saved_positions,
                                pos=pos_drift,
                            )
                            if not found_position:
                                continue

                            # Method of Images for semi-infinite planes
                            if elinfo["type"] == "mea":
                                espikes = espikes * 2
                            espikes = center_espike(espikes, cut_out)
                            drift_spikes.append(espikes)
                            drift_pos.append(pos)

                        # reverse rotation
                        rev_rot = [-r for r in rot]
                        cell.set_rotation(rev_rot[0], rev_rot[1], rev_rot[2], rotation_order="zyx")

                        drift_spikes = np.array(drift_spikes)
                        drift_pos = np.array(drift_pos)
                        if verbose == 2:
                            print(
                                f"Drift done from {np.round(init_pos, 1)} to {np.round(final_pos, 1)} um"
                                f" with {drift_steps} steps"
                            )

                        amp = np.round(np.max(np.abs(drift_spikes[0])), 3)

                        saved_amplitudes.append(amp)
                        saved_eaps.append(drift_spikes)
                        saved_positions.append(drift_pos)
                        saved_rotations.append(rot)
                        if verbose >= 1:
                            print(f"Cell: {cell_name} Progress: [{len(saved_eaps)}/{target_num_spikes}]")
                        saved += 1
                    else:
                        if verbose == 2:
                            print(f"Discarded for trials {cell_name}")
                else:
                    if verbose == 2:
                        print(f"Discarded for minimum amp {cell_name}")
                    pass
            i += 1

            if timeout is not None:
                if time.time() - start_time > timeout:
                    if verbose >= 1:
                        print(f"Timeout finding spikes above noise level for " f"{cell_name}, more than {timeout}")
                    break

    else:
        target_num_spikes = len(cell_locations)
        for loc, rot in zip(cell_locations, cell_rotations):
            # Each cell has several spikes to choose from
            spike_idx = np.random.randint(0, i_spikes.shape[0])
            cell.imem = i_spikes[spike_idx, :, :]
            cell.somav = v_spikes[spike_idx, :]
            cell.set_rotation(rot[0], rot[1], rot[2])

            if not drifting:
                if loc.ndim == 1:
                    pos = loc
                elif loc.ndim == 1:
                    # take first drifting location
                    pos = loc[0]

                espikes, pos_, rot_, found_position = return_extracellular_spike(
                    cell=cell,
                    cell_name=cell_name,
                    model_type=model_type,
                    electrodes=electrodes,
                    limits=[x_lim, y_lim, z_lim],
                    rotation=None,
                    saved_pos=saved_positions,
                    pos=pos,
                )
                if not found_position:
                    continue

                # Method of Images for semi-infinite planes
                if elinfo["type"] == "mea":
                    espikes = espikes * 2

                espikes = center_espike(espikes, cut_out)
                saved_eaps.append(espikes)
                saved_positions.append(pos)
                saved_rotations.append(rot)
            else:
                assert loc.ndim == 2
                drift_spikes = []
                for pos_drift in loc:
                    espikes, pos, r_, found_position = return_extracellular_spike(
                        cell=cell,
                        cell_name=cell_name,
                        model_type=model_type,
                        electrodes=electrodes,
                        limits=[x_lim, y_lim, z_lim],
                        rotation=None,
                        saved_pos=saved_positions,
                        pos=pos_drift,
                    )
                    if not found_position:
                        continue

                    # Method of Images for semi-infinite planes
                    if elinfo["type"] == "mea":
                        espikes = espikes * 2
                    espikes = center_espike(espikes, cut_out)
                    drift_spikes.append(espikes)

                saved_eaps.append(drift_spikes)
                saved_positions.append(loc)
                saved_rotations.append(rot)

            # reverse rotation
            rev_rot = [-r for r in rot]
            cell.set_rotation(rev_rot[0], rev_rot[1], rev_rot[2], rotation_order="zyx")

            if verbose >= 1:
                print(f"Cell: {cell_name} Progress: [{len(saved_eaps)}/{target_num_spikes}]")
            saved += 1
        else:
            pass

    if verbose >= 1:
        print(f"Done generating EAPs for {cell_name}")

    saved_eaps = np.array(saved_eaps, dtype=np.float32)
    saved_positions = np.array(saved_positions)
    saved_rotations = np.array(saved_rotations)

    if save:
        np.save(str(save_folder / f"eap-{cell_save_name}"), saved_eaps)
        np.save(str(save_folder / f"pos-{cell_save_name}"), saved_positions)
        np.save(str(save_folder / f"rot-{cell_save_name}"), saved_rotations)
    else:
        return saved_eaps, saved_positions, saved_rotations


def check_espike(espikes, min_amp, check_shape=True):
    """
    Check extracellular spike amplitude and shape (neg peak > pos peak)

    Parameters
    ----------
    espike: np.array
        EAP (n_elec, n_samples)
    min_amp: float
        Minimum amplitude
    check_shape: bool
        If True, it checks that the minimum peak is larger than the maximum peak

    Returns
    -------
    valid: bool
        If True EAP is valid
    """
    if np.abs(np.min(espikes)) < min_amp:
        return False
    elif check_shape and np.abs(np.min(espikes)) < np.abs(np.max(espikes)):
        return False
    else:
        return True


def center_espike(espike, cut_out_samples, tol=1):
    """
    Centers extracellular spike if the peak is not aligned.

    Parameters
    ----------
    espike : np.array
        EAP (n_elec, n_samples)
    cut_out_samples : list
        Samples before and after the peak
    tol : int
        Tolerance in number of samples

    Returns
    -------
    espike_centered: np.array
        Centered EAP (n_elec, n_samples)
    """
    expexted_peak = cut_out_samples[0]
    peak_idx = np.unravel_index(np.argmin(espike), espike.shape)[1]

    if np.abs(expexted_peak - peak_idx) >= tol:
        if expexted_peak - peak_idx < 0:
            diff = peak_idx - expexted_peak
            cent_espike = np.zeros_like(espike)
            cent_espike[:, :-diff] = espike[:, diff:]
            cent_espike[:, -diff:] = np.tile(espike[:, -1, np.newaxis], [1, diff])
        elif expexted_peak - peak_idx > 0:
            diff = expexted_peak - peak_idx
            cent_espike = np.zeros_like(espike)
            cent_espike[:, diff:] = espike[:, :-diff]
            cent_espike[:, :diff] = np.tile(espike[:, 0, np.newaxis], [1, diff])
    else:
        cent_espike = espike

    cent_peak_idx = np.unravel_index(np.argmin(cent_espike), cent_espike.shape)[1]
    assert np.abs(expexted_peak - cent_peak_idx) < tol, "Something went wrong in centering the spike"

    return cent_espike


def skip_duplicate(pos, saved_positions, drifting, verbose=False):
    """
    Checks if a position has to be skipped because already used.

    Parameters
    ----------
    pos: 3d array
        The position to be tested
    saved_positions: list
        The list of 3d positions already saved
    drifting: bool
        Whether templates are drifting or not
    verbose: bool
        If True, the output is verbose

    Returns
    -------
    skip_duplicate: bool
        If True, the position should be skipped because it's a duplicate
    """
    skip_pos = False
    if len(saved_positions) > 0:
        for pos_s in saved_positions:
            if not drifting:
                test_pos = pos_s
            else:
                test_pos = pos_s[0]
            if np.all(np.round(test_pos, 2) == np.round(pos, 2)):
                if verbose >= 2:
                    print(f"Duplicated position: {np.round(pos, 2)} -- {np.round(pos_s, 2)}. Skipping")
                skip_pos = True
    return skip_pos


def get_physrot_specs(cell_name, model):
    """Return physrot specifications for cell types

    Parameters
    -----------
    cell_name : string
        The name of the cell.
    Returns
    --------
    polarlim : array_like
        lower and upper bound for the polar angle
    pref_orient : array_like
        3-dim vetor of preferred orientation
    """
    if model == "bbp":
        polarlim = {
            "BP": [0.0, 15.0],
            "AC": None,
            "BTC": None,  # [0.,15.],
            "ChC": None,  # [0.,15.],
            "DBC": None,  # [0.,15.],
            "LBC": None,  # [0.,15.],
            "MC": [0.0, 15.0],
            "NBC": None,
            "NGC": None,
            "SBC": None,
            "PC": [0.0, 15.0],
            "SS": [0.0, 15.0],
            "SP": [0.0, 15.0],
        }
        # how it's implemented, the NMC y axis points into the pref_orient direction after rotation
        pref_orient = {
            "BP": [0.0, 0.0, 1.0],
            "AC": None,
            "BTC": None,  # [0.,0.,1.],
            "ChC": None,  # [0.,0.,1.],
            "DBC": None,  # [0.,0.,1.],
            "LBC": None,  # [0.,0.,1.],
            "MC": [0.0, 0.0, 1.0],
            "NBC": None,
            "NGC": None,
            "SBC": None,
            "PC": [0.0, 0.0, 1.0],
            "SS": [0.0, 0.0, 1.0],
            "SP": [0.0, 0.0, 1.0],
        }
        polar = None
        orient = None
        for k in polarlim.keys():
            if k in cell_name.split("_")[1]:
                polar = polarlim[k]
                orient = pref_orient[k]
                break
        return polar, orient
    else:
        raise NotImplementedError("Cell model %s is not implemented" % model)


def return_extracellular_spike(
    cell, cell_name, model_type, electrodes, limits, rotation, saved_pos, pos=None, max_iter=1000, verbose=False
):
    """
    Calculate extracellular spike on MEA at random position relative to cell

    Parameters
    ----------
    cell: LFPy.Cell
        cell object from LFPy
    cell_name: string
        name of cell model
    electrodes: LFPyRecExtElectrode
        The LFPy electrode object
    limits: array_like
        boundaries for neuron locations, shape=(3,2)
    rotation: string
        random rotation to apply to the neuron ('Norot', '3drot', 'physrot')
    saved_pos: np.array
        List of positions already used to be skipped
    pos: array_like, (optional, default None)
        Can be used to set the cell soma to a specific position. If ``None``,
        the random position is used.
    max_iter: int
        Max number of iterations to find a position that hasn't been used yet

    Returns
    -------
    Extracellular spike for each MEA contact site
    """

    def get_xyz_angles(R):
        """Get rotation angles for each axis from rotation matrix

        Parameters;
        -----------
        R : matrix
            3x3 rotation matrix

        Returns
        --------
        R_z : float
        R_y : float
        R_x : float
            Three angles for rotations around axis, defined by R = R_z.R_y.R_x
        """
        rot_x = np.arctan2(R[2, 1], R[2, 2])
        rot_y = np.arcsin(-R[2, 0])
        rot_z = np.arctan2(R[1, 0], R[0, 0])
        return rot_x, rot_y, rot_z

    def get_rnd_rot_Arvo():
        """Generate uniformly distributed random rotation matrices
        see: 'Fast Random Rotation Matrices' by Arvo (1992)

        Returns
        --------
        R : 3x3 matrix
            random rotation matrix
        """
        gamma = np.random.uniform(0, 2.0 * np.pi)
        rotation_z = np.array([[np.cos(gamma), -np.sin(gamma), 0], [np.sin(gamma), np.cos(gamma), 0], [0, 0, 1]])
        x = np.random.uniform(size=2)
        v = np.array(
            [np.cos(2.0 * np.pi * x[0]) * np.sqrt(x[1]), np.sin(2.0 * np.pi * x[0]) * np.sqrt(x[1]), np.sqrt(1 - x[1])]
        )
        H = np.identity(3) - 2.0 * np.outer(v, v)
        M = -np.dot(H, rotation_z)
        return M

    def check_solidangle(matrix, pre, post, polarlim):
        """Check whether a matrix rotates the vector 'pre' into a region
            defined by 'polarlim' around the vector 'post'

        Parameters
        -----------
        matrix : matrix
            3x3 rotation matrix
        pre : array_like
            3-dim vector to be rotated
        post : array_like
            axis of the cones defining the post-rotation region
        polarlim : [float,float]
            Angles specifying the opening of the inner and outer cone
            (aperture = 2*polarlim),
            i.e. the angle between rotated pre vector and post vector has to ly
            within these polar limits.

        Returns
        --------
        test : bool
            True if the vector np.dot(matrix,pre) lies inside the specified region.
        """
        postest = np.dot(matrix, pre)
        c = np.dot(post / np.linalg.norm(post), postest / np.linalg.norm(postest))
        if np.cos(np.deg2rad(polarlim[1])) <= c <= np.cos(np.deg2rad(polarlim[0])):
            return True
        else:
            return False

    # rotate neuron
    if rotation == "norot":
        if model_type == "bbp":
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.0
            y_rot_offset = 0
            z_rot_offset = 0
        else:
            x_rot_offset = 0
            y_rot_offset = 0
            z_rot_offset = 0
        x_rot = x_rot_offset
        y_rot = y_rot_offset
        z_rot = z_rot_offset
    elif rotation == "xrot":
        if model_type == "bbp":
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.0
            y_rot_offset = 0
            z_rot_offset = 0
        else:
            x_rot_offset = 0
            y_rot_offset = 0
            z_rot_offset = 0
        x_rot, _, _ = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot + x_rot_offset
        y_rot = y_rot_offset
        z_rot = z_rot_offset
    elif rotation == "yrot":
        if model_type == "bbp":
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.0
            y_rot_offset = 0
            z_rot_offset = 0
        else:
            x_rot_offset = 0
            y_rot_offset = 0
            z_rot_offset = 0
        _, y_rot, _ = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot_offset
        y_rot = y_rot + y_rot_offset
        z_rot = z_rot_offset
    elif rotation == "zrot":
        if model_type == "bbp":
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.0
            y_rot_offset = 0
            z_rot_offset = 0
        else:
            x_rot_offset = 0
            y_rot_offset = 0
            z_rot_offset = 0
        _, _, z_rot = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot_offset
        y_rot = y_rot_offset
        z_rot = z_rot + z_rot_offset
    elif rotation == "3drot":
        if model_type == "bbp":
            x_rot_offset = np.pi / 2.0  # align neuron with z axis
            y_rot_offset = 0  # align neuron with z axis
            z_rot_offset = 0  # align neuron with z axis
        else:
            x_rot_offset = 0
            y_rot_offset = 0
            z_rot_offset = 0
        x_rot, y_rot, z_rot = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot + x_rot_offset
        y_rot = y_rot + y_rot_offset
        z_rot = z_rot + z_rot_offset
    elif rotation == "physrot":
        polarlim, pref_orient = get_physrot_specs(cell_name, model_type)
        if model_type == "bbp":
            x_rot_offset = np.pi / 2.0  # align neuron with z axis
            y_rot_offset = 0  # align neuron with z axis
            z_rot_offset = 0  # align neuron with z axis
        else:
            raise NotImplementedError("'physrot' rotation is only available with BBP cells")
        while True:
            R = np.array(get_rnd_rot_Arvo())
            if polarlim is None or pref_orient is None:
                valid = True
            else:
                valid = check_solidangle(R, [0.0, 0.0, 1.0], pref_orient, polarlim)
            if valid:
                x_rot, y_rot, z_rot = get_xyz_angles(R)
                x_rot = x_rot + x_rot_offset
                y_rot = y_rot + y_rot_offset
                z_rot = z_rot + z_rot_offset
                break
    else:
        rotation = None
        x_rot = 0
        y_rot = 0
        z_rot = 0

    if pos is None:
        """Move neuron randomly"""
        skip_dup = True
        iter = 0
        x_rands = np.random.uniform(limits[0][0], limits[0][1], 1000)
        y_rands = np.random.uniform(limits[1][0], limits[1][1], 1000)
        z_rands = np.random.uniform(limits[2][0], limits[2][1], 1000)
        while skip_dup and iter < max_iter:
            x_rand = x_rands[iter]
            y_rand = y_rands[iter]
            z_rand = z_rands[iter]
            pos = [x_rand, y_rand, z_rand]
            skip_dup = skip_duplicate(pos, saved_pos, drifting=False, verbose=verbose)
            iter += 1
        if iter == max_iter:
            found_position = False
            return None, None, None, found_position
        else:
            found_position = True
        cell.set_pos(x_rand, y_rand, z_rand)
    else:
        cell.set_pos(pos[0], pos[1], pos[2])
        found_position = True

    cell.set_rotation(x=x_rot, y=y_rot, z=z_rot)
    rot = [x_rot, y_rot, z_rot]

    lfp = np.array(electrodes.get_transformation_matrix() @ cell.imem, dtype=np.float32)

    # Reverse rotation to bring cell back into initial rotation state
    if rotation is not None:
        rev_rot = [-r for r in rot]
        cell.set_rotation(rev_rot[0], rev_rot[1], rev_rot[2], rotation_order="zyx")

    return 1e3 * lfp, pos, rot, found_position


def str2bool(v):
    """Transform string to bool

    Parameters
    -----------
    v : str

    Returns
    --------
    transformed_v, bool
        If v is any of ("yes", "true", "t", "1") (case insensitive)
        ``True`` is returned, else ``False``
    """
    return v.lower() in ("yes", "true", "t", "1")


def simulate_templates_one_cell(cell_model, intra_save_folder, params, verbose, custom_return_cell_function=None):
    """

    Parameters
    ----------
    cell_model
    intra_save_folder
    params
    verbose
    custom_return_cell_function

    Returns
    -------

    """
    run_cell_model(
        cell_model,
        save=True,
        sim_folder=intra_save_folder,
        verbose=verbose,
        custom_return_cell_function=custom_return_cell_function,
        **params,
    )
    print(f"Extracellular simulation: {cell_model}")
    eaps, locs, rots = calc_extracellular(
        0,
        cell_model,
        intra_save_folder,
        verbose=verbose,
        save=False,
        custom_return_cell_function=custom_return_cell_function,
        **params,
    )

    return eaps, locs, rots


def compile_models(cell_folder):
    compile_all_mechanisms(cell_folder)
    print(f"Compiled all cell models in {cell_folder}")


def compute_eap_for_cell_model(i, cell_model, params_path, intraonly=False, verbose=False):
    params = safe_yaml_load(params_path)

    extra_sim_folder = Path(params["templates_folder"])
    vm_im_sim_folder = Path(params["templates_folder"]) / "intracellular"

    print(f"Intracellular simulation: {cell_model}")
    run_cell_model(cell_model_folder=cell_model, save=True, sim_folder=vm_im_sim_folder, verbose=verbose, **params)
    if not intraonly:
        print(f"Extracellular simulation: {cell_model}")
        calc_extracellular(
            i,
            cell_model_folder=cell_model,
            save_sim_folder=extra_sim_folder,
            load_sim_folder=vm_im_sim_folder,
            verbose=verbose,
            **params,
        )


def compute_eap_based_on_tempgen(cell_folder, params_path, tempgen, intraonly=False, verbose=False):
    params = safe_yaml_load(params_path)

    extra_sim_folder = params["templates_folder"]
    vm_im_sim_folder = str(Path(params["templates_folder"]) / "intracellular")

    celltypes = np.unique(tempgen.celltypes)

    for celltype in celltypes:
        celltype_idxs = np.where(tempgen.celltypes == celltype)
        if np.any(np.diff(celltype_idxs) != 1):
            raise NotImplementedError("Cell types in the template generator must be contiguous.")

    # print(f'Intracellular simulation: {cell_model}')
    for i, celltype in enumerate(celltypes):
        cell_model = cell_folder / celltype
        run_cell_model(cell_model_folder=cell_model, sim_folder=vm_im_sim_folder, verbose=verbose, **params)
        celltype_idxs = np.where(tempgen.celltypes == celltype)
        cell_locations = tempgen.locations[celltype_idxs]
        cell_rotations = tempgen.rotations[celltype_idxs]
        if not intraonly:
            print(f"Extracellular simulation: {cell_model}")
            calc_extracellular(
                i,
                cell_model_folder=cell_model,
                save_sim_folder=extra_sim_folder,
                load_sim_folder=vm_im_sim_folder,
                verbose=verbose,
                cell_locations=cell_locations,
                cell_rotations=cell_rotations,
                **params,
            )


if __name__ == "__main__":
    if len(sys.argv) == 3 and sys.argv[1] == "compile":
        cell_folder = sys.argv[2]
        compile_all_mechanisms(cell_folder)
        sys.exit(0)
    elif len(sys.argv) == 6:
        i = int(sys.argv[1])
        cell_model = sys.argv[2]
        intraonly = str2bool(sys.argv[3])
        params_path = sys.argv[4]
        verbose = int(sys.argv[5])
        compute_eap_for_cell_model(i, cell_model, params_path, intraonly=intraonly, verbose=verbose)



import numpy as np
import scipy.signal


def generate_drift_dict_from_params(
    drift_fs,
    duration,
    template_locations,
    external_drift_vector_um=None,
    external_drift_times=None,
    t_start_drift=None,
    t_end_drift=None,
    drift_mode_probe="rigid",
    drift_mode_speed="slow",
    non_rigid_gradient_mode="linear",
    non_rigid_linear_direction=1,
    non_rigid_linear_min_factor=0.5,
    non_rigid_step_depth_boundary=None,
    non_rigid_step_factors=None,
    preferred_dir=[0, 0, 1],
    slow_drift_velocity=5,
    slow_drift_waveform="triangluar",
    slow_drift_amplitude=None,
    fast_drift_period=10.0,
    fast_drift_max_jump=20.0,
    fast_drift_min_jump=5.0,
    external_drift_factors=None,
):
    """
    Generate for each units a drift position vector.
    The output vector handle  of indexes for drift steps along time.

    Parameters
    ----------
    drift_fs: float
        Sampling frequency in Hz
    duration: float
        Duration drift vector in s
    template_locations: np.array 3d
        Shape: (num_cells, drift_steps, 3)
        Template locations for every drift steps
    external_drift_vector_um : 1d array or None
        External drift signal to apply. It needs to be consistent with drift_fs and duration (or external_drift_times)
    external_drift_times : 1d array or None
        External drift times to use for external_drift_signal_um.
    t_start_drift: float
        When the drift start in second
    drift_mode_probe: 'rigid' or 'non-rigid'
        Global drift or per cell.
    drift_mode_speed: str 'slow', 'fast', 'slow+fast'
        Drift mode speed
    non_rigid_gradient_mode: 'linear' | 'step'
        Mode to generate non rigid gradient over units on velocity.
        If 'linear', drift factors (0.5-1) are assigned to each cells linearly with depth (deeper cells have more drift).
    non_rigid_linear_direction : int
        Direction of linear gradient.
        Only used if 'drift_mode_probe' is 'non-rigid' and 'non_rigid_gradient_mode' is 'linear'
        If positive, deeper cells drift more than surface cells.
        If negative, deeper cells drift more than surface cells.
        Zero is not allowed, by default 1
    non_rigid_linear_min_factor: float (default 0.5)
        When non_rigid_gradient_mode='linear' this control the minimum factor. The max is always 1.
    non_rigid_step_depth_boundary : float or None
        Depth boundary in the preferred_dir dimension to apply the drift step,
        by default is half of the template locations extensions.
        Only used if 'drift_mode_probe' is 'non-rigid' and 'non_rigid_gradient_mode' is 'step'
    non_rigid_step_factors : None or tuple/list of two elements
        Factors to apply to cells above (non_rigid_step_factors[0]) and below (non_rigid_step_factors[1])
        step depth boundary. Default is (1, 0).
        Only used if 'drift_mode_probe' is 'non-rigid' and 'non_rigid_gradient_mode' is 'step'
    preferred_dir: list of int
        Use for gradient when non rigid mode.
    slow_drift_velocity: float
        Slow drift velocity in um/min.
    slow_drift_waveform: 'trianglar' or 'sine'
        Waveform shape for slow drift
    fast_drift_period: float
        Period between fast drift events
    fast_drift_max_jump: float
        Maximum 'jump' in um for fast drifts
    fast_drift_min_jump: float
        Minimum 'jump' in um for fast drifts
    external_drift_factors: None
        If templates are selected externally (and template_ids) is used in the
        generate_recordings() function, this vector specifies the factor (0-1)
        for each cell.

    Returns
    -------
    drift_list: list of dicts
        Each dict contains info for a drift signal:
            * "drift_times" (1d array): times for different drift vectors. (Only for external times)
            * "drift_fs" (float): sampling frequency
            * "drift_vector_um" (1d array): drift signals in um
            * "drift_vector_idxs" (1d array): drift signals in template idxs (with respect to middle step)
            * "drift_factor" (1d array): array with gradient for each unit for the drift signal
               vector. For rigid motion, all values are 1.
    """

    num_cells, drift_steps, _ = template_locations.shape
    n_samples = int(drift_fs * duration)

    drift_distances = np.linalg.norm(template_locations[:, -1] - template_locations[:, 0])
    mean_distance = np.mean(drift_distances)
    assert np.max(drift_distances - mean_distance) < 0.05 * mean_distance, (
        "Drift distances are not constant across templates. You can control this by setting "
        "'max_drift' and 'min_drift' to the same value and 'drift_zlim = [desired_drift, desired_drift]' "
        "in the template parameters."
    )
    # velocity and jump are computed on bigger drift over cell
    loc0 = template_locations[:, 0, :]
    loc1 = template_locations[:, -1, :]
    dist = np.sum((loc0 - loc1) ** 2, axis=1) ** 0.5
    dist_boundary = np.min(dist)
    step = dist_boundary / drift_steps

    if external_drift_vector_um is None:
        assert drift_mode_speed in ["slow", "fast"]
        t_start_drift = 0.0 if t_start_drift is None else t_start_drift
        t_end_drift = duration if t_end_drift is None else t_end_drift
        assert t_start_drift < duration, f"'t_start_drift' must preceed 'duration'!"
        assert t_end_drift <= duration, f"'t_end_drift' must preceed 'duration'!"
        start_drift_index = int(t_start_drift * drift_fs)
        end_drift_index = int(t_end_drift * drift_fs)

        drift_vector_um = np.zeros(n_samples, dtype="float32")
        # drift times is only used for external drift that are not defined over the entire recording
        drift_times = None

        if drift_mode_speed == "slow":
            if slow_drift_amplitude is None:
                slow_drift_amplitude = dist_boundary

            # compute half period for a regular drift speed
            half_period = slow_drift_amplitude / (slow_drift_velocity / 60)

            # triangle / sine frequency depends on the velocity
            freq = 1.0 / (2 * half_period)

            drift_times = np.arange(end_drift_index - start_drift_index) / drift_fs

            if slow_drift_waveform == "triangluar":
                triangle = np.abs(scipy.signal.sawtooth(2 * np.pi * freq * drift_times + np.pi / 2))
                triangle *= slow_drift_amplitude
                triangle -= slow_drift_amplitude / 2.0

                drift_vector_um[start_drift_index:end_drift_index] = triangle
                drift_vector_um[end_drift_index:] = triangle[-1]
            elif slow_drift_waveform == "sine":
                sine = np.cos(2 * np.pi * freq * drift_times + np.pi / 2)
                sine *= slow_drift_amplitude / 2.0

                drift_vector_um[start_drift_index:end_drift_index] = sine
                drift_vector_um[end_drift_index:] = sine[-1]
            else:
                raise NotImplementedError("slow_drift_waveform")

        else:  # 'fast' in drift_mode_speed:
            period_samples = int(fast_drift_period * drift_fs)
            n = int(np.round((end_drift_index - start_drift_index) / period_samples))

            t_idx = start_drift_index + period_samples
            for i in range(1, n):
                jump = np.random.rand() * (fast_drift_max_jump - fast_drift_min_jump) + fast_drift_min_jump
                if np.random.rand() > 0.5:
                    jump = -jump

                # protect from boundaries
                if jump > 0:
                    if (np.max(drift_vector_um[t_idx : t_idx + period_samples]) + jump) >= dist_boundary / 2.0:
                        jump = -jump

                # protect from boundaries
                if jump < 0:
                    if (np.min(drift_vector_um[t_idx : t_idx + period_samples]) + jump) <= -dist_boundary / 2.0:
                        jump = -jump

                drift_vector_um[t_idx:] += jump
                t_idx += period_samples
    else:
        if external_drift_times is None:
            assert (
                len(external_drift_vector_um) == n_samples
            ), f"'external_drift_times' must have a length of {n_samples} (duration * drift_fs)."
        else:
            assert all(np.diff(external_drift_times) > 0), "'external_drift_times' should be monotonically increasing"
            assert external_drift_times[-1] <= duration, "'external_drift_times' cannot exceed duration"
            assert len(external_drift_times) == len(
                external_drift_vector_um
            ), "'external_drift_times' and 'external_drift_vector_um' must have the same length"
            drift_vector_um = external_drift_vector_um
            drift_vector_um *= 0.99999
            drift_times = external_drift_times

    if external_drift_factors is None:
        if drift_mode_probe == "rigid":
            drift_factors = np.ones(num_cells)
        elif drift_mode_probe == "non-rigid":
            assert non_rigid_gradient_mode in ("linear", "step")
            # project locations over preferred direction
            preferred_dir = np.array(preferred_dir).reshape(-1, 1)
            locs = template_locations[:, drift_steps // 2, :]
            proj = np.dot(locs, preferred_dir).squeeze()
            if non_rigid_gradient_mode == "linear":
                assert 0 <= non_rigid_linear_min_factor < 1, "non_rigid_linear_min_factor must between 0 and 1"
                assert non_rigid_linear_direction in (1, -1, 1.0, -1.0), "non_rigid_linear_direction must be 1 or -1"
                # vector with shape (num_cells, ) and value between 0.5 and 1 which is a factor of the velocity
                # the upper units on the 'preferred_dir' vector get 0.5 the lower get 1
                print(f"Linear gradient with Min: {non_rigid_linear_min_factor} Dir: {non_rigid_linear_direction}")
                drift_factors = (proj - np.min(proj)) / (np.max(proj) - np.min(proj))
                if non_rigid_linear_direction > 0:
                    drift_factors = 1 - drift_factors
                f = non_rigid_linear_min_factor
                drift_factors = drift_factors * (1 - f) + f
            elif non_rigid_gradient_mode == "step":
                if non_rigid_step_depth_boundary is None:
                    non_rigid_step_depth_boundary = np.mean(proj)
                else:
                    assert (
                        np.min(proj) < non_rigid_step_depth_boundary < np.max(proj)
                    ), f"'non_rigid_step_depth_boundary' needs to be in between {np.min(proj)} and {np.max(proj)}"
                if non_rigid_step_factors is None:
                    non_rigid_step_factors = (1, 0)
                else:
                    assert len(non_rigid_step_factors) == 2, "'non_rigid_step_factors' needs two values"
                    assert (
                        0 <= non_rigid_step_factors[0] <= 1
                    ), "'non_rigid_step_factors' needs to be in the [0, 1] range"
                    assert (
                        0 <= non_rigid_step_factors[1] <= 1
                    ), "'non_rigid_step_factors' needs to be in the [0, 1] range"
                drift_factors = non_rigid_step_factors[0] * np.ones(num_cells)
                drift_factors[proj < non_rigid_step_depth_boundary] = non_rigid_step_factors[1]
    else:
        assert (
            len(external_drift_factors) == num_cells
        ), f"'external_drift_factors' must be defined for all {num_cells} cells"
        assert all(0 <= d <= 1 for d in external_drift_factors), "'external_drift_factors' must be in the [0, 1] range"
        drift_factors = external_drift_factors

    # avoid possible clipping
    drift_vector_um = np.clip(drift_vector_um, -dist_boundary / 2, dist_boundary / 2)

    # avoid boundary effect
    drift_vector_um *= 0.99999

    # shift with respect to mid point in int16
    drift_vector_idxs = (np.floor(drift_vector_um / step)).astype("int16")

    drift_dict = {}
    drift_dict["drift_vector_um"] = drift_vector_um
    drift_dict["drift_vector_idxs"] = drift_vector_idxs
    drift_dict["drift_fs"] = drift_fs
    drift_dict["drift_times"] = drift_times
    drift_dict["drift_factors"] = drift_factors
    drift_dict["drift_steps"] = drift_steps

    return drift_dict


def test_generate_drift_position_vector():
    fs = 2.0
    duration = 600
    num_cells = 4
    drift_steps = 21
    template_locations = np.zeros((num_cells, drift_steps, 3))

    # one cells every 30 um
    template_locations[:, :, 2] = np.arange(num_cells).reshape(-1, 1) * 25.0

    # drift step 1.5 um of Z
    template_locations[:, :, 2] += np.arange(drift_steps) * 1.5

    drift_dict = generate_drift_dict_from_params(
        drift_fs=fs,
        template_locations=template_locations,
        duration=duration,
        t_start_drift=10.0,
        # drift_mode_probe='rigid',
        drift_mode_probe="non-rigid",
        # drift_mode_speed='slow+fast',
        # drift_mode_speed='fast',
        drift_mode_speed="slow",
        non_rigid_gradient_mode="linear",
        preferred_dir=[0, 0, 1],
        # slow_drift_velocity=5.,
        # slow_drift_waveform='triangluar',
        slow_drift_velocity=17,
        slow_drift_waveform="sine",
        fast_drift_period=120.0,
        fast_drift_max_jump=5,
        fast_drift_min_jump=1.0,
    )

    drift_vectors = drift_dict["drift_vectors"]
    drift_fs = drift_dict["drift_fs"]
    times = np.arange(int(fs * duration)) / drift_fs

    print(drift_vectors.shape)

    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()

    ax.plot(times, drift_vectors)
    ax.axhline(drift_steps - 1, ls="--", color="k")
    plt.ion()
    plt.show()


if __name__ == "__main__":
    test_generate_drift_position_vector()



import shutil
import tempfile
import unittest
from copy import deepcopy
from pathlib import Path

import numpy as np
import yaml
from click.testing import CliRunner
from packaging.version import parse

import MEArec as mr
from MEArec.cli import cli

DEBUG = False


if DEBUG:
    import matplotlib.pyplot as plt

    plt.ion()
    plt.show()


class TestGenerators(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        info, info_folder = mr.get_default_config()
        cell_models_folder = mr.get_default_cell_models_folder()
        self.num_cells = len(
            [f for f in Path(cell_models_folder).iterdir() if "mods" not in f.name and not f.name.startswith(".")]
        )
        self.n = 10
        self.n_drift = 5

        templates_params = mr.get_default_templates_params()
        rec_params = mr.get_default_recordings_params()

        # Set seed
        np.random.seed(2308)

        if not DEBUG:
            self.test_dir = Path(tempfile.mkdtemp())
        else:
            self.test_dir = Path("./tmp").absolute()

        if not (self.test_dir / "templates.h5").is_file() and not (self.test_dir / "templates_drift.h5").is_file():
            templates_params["n"] = self.n
            templates_params["ncontacts"] = 1
            templates_params["probe"] = "Neuronexus-32"
            templates_folder = info["templates_folder"]
            print("Generating non-drifting templates")
            templates_params["min_amp"] = 10
            print(templates_params)
            self.tempgen = mr.gen_templates(
                cell_models_folder,
                templates_tmp_folder=templates_folder,
                params=templates_params,
                parallel=True,
                delete_tmp=True,
                verbose=False,
            )
            self.templates_params = deepcopy(templates_params)
            self.num_templates, self.num_chan, self.num_samples = self.tempgen.templates.shape

            templates_params["drifting"] = True
            templates_params["n"] = self.n_drift
            templates_params["drift_steps"] = 10
            templates_params["rot"] = "norot"
            templates_params["min_amp"] = 10
            print("Generating drifting templates")
            self.tempgen_drift = mr.gen_templates(
                cell_models_folder,
                templates_tmp_folder=templates_folder,
                params=templates_params,
                parallel=True,
                delete_tmp=True,
                verbose=False,
            )
            self.templates_params_drift = deepcopy(templates_params)
            self.num_steps_drift = self.tempgen_drift.templates.shape[1]

            print("Making test recordings to test load functions")
            mr.save_template_generator(self.tempgen, self.test_dir / "templates.h5")
            mr.save_template_generator(self.tempgen_drift, self.test_dir / "templates_drift.h5")
        else:
            print(self.test_dir / "templates.h5")
            self.tempgen = mr.load_templates(self.test_dir / "templates.h5", return_h5_objects=False)
            self.tempgen_drift = mr.load_templates(self.test_dir / "templates_drift.h5")
            self.templates_params = self.tempgen.info["params"]
            self.templates_params_drift = self.tempgen_drift.info["params"]
            self.num_steps_drift = self.tempgen_drift.templates.shape[1]
            self.num_templates, self.num_chan, self.num_samples = self.tempgen.templates.shape

        if not (self.test_dir / "recordings.h5").is_file():
            ne = 2
            ni = 1
            rec_params["spiketrains"]["n_exc"] = ne
            rec_params["spiketrains"]["n_inh"] = ni
            rec_params["recordings"]["modulation"] = "none"
            rec_params["recordings"]["filter"] = False
            rec_params["recordings"]["extract_waveforms"] = True
            rec_params["recordings"]["overlap"] = True
            rec_params["spiketrains"]["duration"] = 2
            rec_params["templates"]["min_dist"] = 1

            self.recgen = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)
            self.recgen.annotate_overlapping_spikes(parallel=False)
            mr.save_recording_generator(self.recgen, self.test_dir / "recordings.h5")
        else:
            self.recgen = mr.load_recordings(self.test_dir / "recordings.h5", return_h5_objects=False)

    @classmethod
    def tearDownClass(self):
        # Remove the directory after the test
        if not DEBUG:
            shutil.rmtree(self.test_dir)

    def test_gen_templates(self):
        print("Test templates generation")
        n = self.n
        num_cells = self.num_cells
        templates_params = self.templates_params

        assert self.tempgen.templates.shape[0] == (n * num_cells)
        assert len(self.tempgen.locations) == (n * num_cells)
        assert len(self.tempgen.rotations) == (n * num_cells)
        assert len(self.tempgen.celltypes) == (n * num_cells)
        assert len(np.unique(self.tempgen.celltypes)) == num_cells
        assert (
            np.min(self.tempgen.locations[:, 0]) > templates_params["xlim"][0]
            and np.max(self.tempgen.locations[:, 0]) < templates_params["xlim"][1]
        )

    def test_gen_templates_drift(self):
        print("Test drifting templates generation")
        n = self.n_drift
        num_cells = self.num_cells
        n_steps = self.num_steps_drift

        assert self.tempgen_drift.templates.shape[0] == (n * num_cells)
        assert self.tempgen_drift.locations.shape == (n * num_cells, n_steps, 3)
        assert len(self.tempgen_drift.rotations) == (n * num_cells)
        assert len(self.tempgen_drift.celltypes) == (n * num_cells)
        assert len(np.unique(self.tempgen_drift.celltypes)) == num_cells
        assert self.tempgen_drift.templates.shape[1] == self.num_steps_drift

    def test_gen_templates_from_tempgen(self):
        print("Test templates generation from existing template generator")
        cell_models_folder = mr.get_default_cell_models_folder()

        # no drift
        params = self.templates_params
        params["probe"] = "Neuropixels-24"
        tempgen2 = mr.gen_templates(cell_models_folder=cell_models_folder, params=params, tempgen=self.tempgen)

        assert tempgen2.templates.shape[0] == self.tempgen.templates.shape[0]
        # assert that all locations are the same
        for loc_new, loc_old in zip(tempgen2.locations, self.tempgen.locations):
            assert np.allclose(loc_new, loc_old)

        # drift
        params = self.templates_params_drift
        params["probe"] = "Neuropixels-24"
        tempgen_drift2 = mr.gen_templates(
            cell_models_folder=cell_models_folder, params=params, tempgen=self.tempgen_drift
        )

        assert tempgen_drift2.templates.shape[0] == self.tempgen_drift.templates.shape[0]
        # assert that all locations are the same
        for loc_new, loc_old in zip(tempgen2.locations, self.tempgen.locations):
            for loc_new_d, loc_old_d in zip(loc_new, loc_old):
                assert np.allclose(loc_new_d, loc_old_d)

    def test_gen_templates_beta_distr(self):
        print("Test templates generation from beta distributions")
        cell_models_folder = mr.get_default_cell_models_folder()

        # no drift
        params = self.templates_params
        params["probe"] = "Neuronexus-32"

        # beta distr
        params["x_distr"] = "beta"
        tempgen_beta = mr.gen_templates(cell_models_folder=cell_models_folder, params=params)

        assert tempgen_beta.templates.shape[0] == self.tempgen.templates.shape[0]

        # drift
        params = self.templates_params_drift
        params["probe"] = "Neuronexus-32"
        params["x_distr"] = "beta"
        params["beta_distr_params"] = [2, 5]

        tempgen_drift_beta = mr.gen_templates(cell_models_folder=cell_models_folder, params=params)

        assert tempgen_drift_beta.templates.shape[0] == self.tempgen_drift.templates.shape[0]

    def test_gen_spiketrains(self):
        import elephant.statistics as stat
        
        print("Test spike train generation")
        rec_params = mr.get_default_recordings_params()
        sp_params = rec_params["spiketrains"]
        spgen = mr.SpikeTrainGenerator(sp_params, seed=0)
        spgen.generate_spikes()

        # check ref period
        for st in spgen.spiketrains:
            isi = stat.isi(st).rescale("ms")
            assert np.all(isi.magnitude > sp_params["ref_per"])
            assert (1 / np.mean(isi.rescale("s"))) > sp_params["min_rate"]

        sp_params["process"] = "gamma"
        spgen = mr.SpikeTrainGenerator(sp_params, seed=0)
        spgen.generate_spikes()
        for st in spgen.spiketrains:
            isi = stat.isi(st).rescale("ms")
            assert np.all(isi.magnitude > sp_params["ref_per"])
            assert (1 / np.mean(isi.rescale("s"))) > sp_params["min_rate"]

        spgen = mr.gen_spiketrains(sp_params, seed=0)
        spiketrains = spgen.spiketrains
        spgen_st = mr.gen_spiketrains(spiketrains=spiketrains)
        for st, st_ in zip(spgen.spiketrains, spgen_st.spiketrains):
            assert np.allclose(st.times.magnitude, st_.times.magnitude)

    def test_gen_recordings_with_spiketrains(self):
        print("Test recording generation - from spiketrains")
        rec_params = mr.get_default_recordings_params()
        sp_params = rec_params["spiketrains"]
        sp_params["n_exc"] = 2
        sp_params["n_inh"] = 1
        sp_params["duration"] = 5
        spgen = mr.gen_spiketrains(sp_params)
        recgen = mr.gen_recordings(params=rec_params, spgen=spgen, tempgen=self.tempgen, verbose=False)

        for st, st_ in zip(spgen.spiketrains, recgen.spiketrains):
            assert np.allclose(st.times.magnitude, st_.times.magnitude)
        del recgen

    def test_gen_recordings_modulations(self):
        print("Test recording generation - modulation")
        ne = 1
        ni = 2
        fe = 30
        fi = 50
        num_chan = self.num_chan
        n_neurons = ne + ni

        rec_params = mr.get_default_recordings_params()

        rec_params["spiketrains"]["n_exc"] = ne
        rec_params["spiketrains"]["n_inh"] = ni
        rec_params["spiketrains"]["f_exc"] = fe
        rec_params["spiketrains"]["f_inh"] = fi
        rec_params["spiketrains"]["duration"] = 3
        n_jitter = [1, 5]
        modulations = ["none", "electrode", "template"]
        chunk_rec = [0, 2]

        rec_params["templates"]["min_dist"] = 1
        bursting = [False, True]

        for mod in modulations:
            for b in bursting:
                for j in n_jitter:
                    for ch in chunk_rec:
                        print("Modulation: modulation", mod, "bursting", b, "jitter", j, "chunk", ch)
                        rec_params["templates"]["n_jitters"] = j
                        rec_params["recordings"]["modulation"] = mod
                        rec_params["recordings"]["bursting"] = b
                        rec_params["recordings"]["chunk_duration"] = ch

                        if mod == "electrode" and b is True and j == 5:
                            rec_params["cell_types"] = None

                        recgen_mod = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

                        assert recgen_mod.recordings.shape[1] == num_chan
                        assert len(recgen_mod.spiketrains) == n_neurons
                        assert recgen_mod.channel_positions.shape == (num_chan, 3)
                        if j == 1:
                            assert recgen_mod.templates.shape[:2] == (n_neurons, num_chan)
                        else:
                            assert recgen_mod.templates.shape[:3] == (n_neurons, j, num_chan)
                        assert recgen_mod.voltage_peaks.shape == (n_neurons, num_chan)
                        assert recgen_mod.spike_traces.shape[1] == n_neurons
                        del recgen_mod

    def test_gen_recordings_bursting(self):
        print("Test recording generation - shape_mod")
        rates = [30, 50]
        types = ["E", "I"]
        num_chan = self.num_chan
        n_neurons = len(rates)

        rec_params = mr.get_default_recordings_params()

        rec_params["spiketrains"]["rates"] = rates
        rec_params["spiketrains"]["types"] = types
        rec_params["spiketrains"]["duration"] = 3
        n_jitter = 4
        rec_params["templates"]["n_jitters"] = n_jitter
        rec_params["recordings"]["modulation"] = "electrode"
        rec_params["recordings"]["bursting"] = True
        rec_params["recordings"]["shape_mod"] = True
        rec_params["recordings"]["sync_rate"] = 0.2
        rec_params["recordings"]["overlap"] = True
        rec_params["recordings"]["extract_waveforms"] = True
        rec_params["templates"]["min_dist"] = 1
        rec_params["templates"]["min_amp"] = 0

        recgen_burst = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)
        recgen_burst.extract_waveforms()

        assert recgen_burst.recordings.shape[1] == num_chan
        assert len(recgen_burst.spiketrains) == n_neurons
        assert recgen_burst.channel_positions.shape == (num_chan, 3)
        assert recgen_burst.templates.shape[:3] == (n_neurons, n_jitter, num_chan)
        assert recgen_burst.voltage_peaks.shape == (n_neurons, num_chan)
        assert recgen_burst.spike_traces.shape[1] == n_neurons
        del recgen_burst

        rec_params["recordings"]["modulation"] = "template"
        rec_params["recordings"]["n_bursting"] = 2

        recgen_burst = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)
        recgen_burst.extract_waveforms()

        assert recgen_burst.recordings.shape[1] == num_chan
        assert len(recgen_burst.spiketrains) == n_neurons
        assert recgen_burst.channel_positions.shape == (num_chan, 3)
        assert recgen_burst.templates.shape[:3] == (n_neurons, n_jitter, num_chan)
        assert recgen_burst.voltage_peaks.shape == (n_neurons, num_chan)
        assert recgen_burst.spike_traces.shape[1] == n_neurons
        del recgen_burst

    def test_gen_recordings_multiple_bursting_modulation(self):
        print("Test recording generation - multiple bursting modulation")
        rates = [10, 20, 30]
        types = ["E", "E", "I"]
        num_chan = self.num_chan
        n_neurons = len(rates)

        rec_params = mr.get_default_recordings_params()

        rec_params["spiketrains"]["rates"] = rates
        rec_params["spiketrains"]["types"] = types
        rec_params["spiketrains"]["duration"] = 3
        n_jitter = 4
        rec_params["templates"]["n_jitters"] = n_jitter
        rec_params["recordings"]["modulation"] = "electrode"
        rec_params["recordings"]["bursting"] = True
        rec_params["recordings"]["shape_mod"] = True
        rec_params["recordings"]["sync_rate"] = 0.2
        rec_params["recordings"]["overlap"] = True
        rec_params["recordings"]["extract_waveforms"] = True
        rec_params["templates"]["min_dist"] = 1
        rec_params["templates"]["min_amp"] = 0

        rec_params["recordings"]["exp_decay"] = [0.1, 0.15, 0.2]
        rec_params["recordings"]["max_burst_duration"] = [50, 100, 150]
        rec_params["recordings"]["n_burst_spikes"] = [5, 10, 15]
        rec_params["recordings"]["shape_stretch"] = 10

        recgen_burst = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)
        recgen_burst.extract_waveforms()

        assert recgen_burst.recordings.shape[1] == num_chan
        assert len(recgen_burst.spiketrains) == n_neurons
        assert recgen_burst.channel_positions.shape == (num_chan, 3)
        assert recgen_burst.templates.shape[:3] == (n_neurons, n_jitter, num_chan)
        assert recgen_burst.voltage_peaks.shape == (n_neurons, num_chan)
        assert recgen_burst.spike_traces.shape[1] == n_neurons

        for st in recgen_burst.spiketrains:
            assert st.annotations["bursting"]
        del recgen_burst

        rec_params["recordings"]["modulation"] = "template"
        rec_params["recordings"]["bursting_units"] = [1, 2]
        rec_params["recordings"]["exp_decay"] = [0.1, 0.15]
        rec_params["recordings"]["max_burst_duration"] = [50, 100]
        rec_params["recordings"]["n_burst_spikes"] = [10, 15]
        rec_params["recordings"]["shape_stretch"] = 20

        recgen_burst = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)
        recgen_burst.extract_waveforms()

        assert recgen_burst.recordings.shape[1] == num_chan
        assert len(recgen_burst.spiketrains) == n_neurons
        assert recgen_burst.channel_positions.shape == (num_chan, 3)
        assert recgen_burst.templates.shape[:3] == (n_neurons, n_jitter, num_chan)
        assert recgen_burst.voltage_peaks.shape == (n_neurons, num_chan)
        assert recgen_burst.spike_traces.shape[1] == n_neurons
        for i, st in enumerate(recgen_burst.spiketrains):
            if i in rec_params["recordings"]["bursting_units"]:
                assert st.annotations["bursting"]
            else:
                assert not st.annotations["bursting"]
        del recgen_burst

    def test_gen_recordings_noise(self):
        print("Test recording generation - noise")
        ne = 0
        ni = 0
        num_chan = self.num_chan
        n_neurons = ne + ni

        rec_params = mr.get_default_recordings_params()

        rec_params["spiketrains"]["n_exc"] = ne
        rec_params["spiketrains"]["n_inh"] = ni
        rec_params["spiketrains"]["duration"] = 3
        n_jitter = 3
        rec_params["templates"]["n_jitters"] = n_jitter
        rec_params["templates"]["min_dist"] = 1
        rec_params["recordings"]["modulation"] = "none"
        noise_modes = ["uncorrelated", "distance-correlated"]
        chunk_noise = [0, 2]
        noise_color = [True, False]
        noise_level = 10
        rec_params["recordings"]["noise_level"] = noise_level
        rec_params["recordings"]["filter"] = False

        for mode in noise_modes:
            for ch in chunk_noise:
                for color in noise_color:
                    print("Noise: mode", mode, "chunks", ch, "color", color)
                    rec_params["recordings"]["noise_mode"] = mode
                    rec_params["recordings"]["chunk_duration"] = ch
                    rec_params["recordings"]["noise_color"] = color
                    recgen_noise = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

                    if mode == "uncorrelated" and ch == 0 and not noise_color:
                        rec_params["recordings"]["fs"] = 30000
                        rec_params["recordings"]["chunk_duration"] = 1
                    if noise_color:
                        rec_params["recordings"]["filter_cutoff"] = 500

                    assert recgen_noise.recordings.shape[1] == num_chan
                    assert len(recgen_noise.spiketrains) == n_neurons
                    assert recgen_noise.channel_positions.shape == (num_chan, 3)
                    assert len(recgen_noise.spike_traces) == n_neurons
                    assert np.isclose(np.std(recgen_noise.recordings), noise_level, atol=1)
                    del recgen_noise

    def test_gen_recordings_far_neurons(self):
        print("Test recording generation - far neurons")
        ne = 0
        ni = 0
        num_chan = self.num_chan
        n_neurons = ne + ni

        rec_params = mr.get_default_recordings_params()

        rec_params["spiketrains"]["n_exc"] = ne
        rec_params["spiketrains"]["n_inh"] = ni
        rec_params["spiketrains"]["duration"] = 3
        n_jitter = 3
        rec_params["templates"]["n_jitters"] = n_jitter
        rec_params["templates"]["min_dist"] = 1
        rec_params["recordings"]["modulation"] = "none"
        rec_params["recordings"]["noise_mode"] = "far-neurons"
        rec_params["recordings"]["far_neurons_n"] = 10
        rec_params["recordings"]["far_neurons_max_amp"] = 100
        rec_params["recordings"]["far_neurons_exc_inh_ratio"] = 0.8
        noise_level = 20
        rec_params["recordings"]["noise_level"] = noise_level
        rec_params["recordings"]["filter"] = False
        recgen_noise = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

        assert recgen_noise.recordings.shape[1] == num_chan
        assert recgen_noise.channel_positions.shape == (num_chan, 3)
        assert len(recgen_noise.spiketrains) == n_neurons
        assert len(recgen_noise.spike_traces) == n_neurons
        assert np.isclose(np.std(recgen_noise.recordings), noise_level, atol=1)
        del recgen_noise

        rec_params["recordings"]["chunk_duration"] = 1
        recgen_noise = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

        assert recgen_noise.recordings.shape[1] == num_chan
        assert recgen_noise.channel_positions.shape == (num_chan, 3)
        assert len(recgen_noise.spiketrains) == n_neurons
        assert len(recgen_noise.spike_traces) == n_neurons
        assert np.isclose(np.std(recgen_noise.recordings), noise_level, atol=1)
        del recgen_noise

    def test_gen_recordings_filters(self):
        print("Test recording generation - filters")
        ne = 1
        ni = 1
        num_chan = self.num_chan
        n_neurons = ne + ni

        rec_params = mr.get_default_recordings_params()

        filter_modes = ["filtfilt", "lfilter"]
        filter_orders = [1, 3, 5]
        filter_cutoffs = [300, 500, [300, 3000], [300, 6000]]
        chunk_rec = [0, 2]

        rec_params["spiketrains"]["n_exc"] = ne
        rec_params["spiketrains"]["n_inh"] = ni
        rec_params["spiketrains"]["duration"] = 5
        rec_params["recordings"]["filter"] = True
        rec_params["templates"]["min_dist"] = 1

        for mode in filter_modes:
            for order in filter_orders:
                for cutoff in filter_cutoffs:
                    for ch in chunk_rec:
                        print(f"Filter: mode {mode} order {order} cutoff {cutoff} chunk dur {ch}")
                        rec_params["recordings"]["chunk_duration"] = ch
                        rec_params["recordings"]["filter_mode"] = mode
                        rec_params["recordings"]["filter_order"] = order
                        rec_params["recordings"]["filter_cutoff"] = cutoff

                        recgen_filt = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)
                        assert recgen_filt.recordings.shape[1] == num_chan
                        assert len(recgen_filt.spiketrains) == n_neurons
                        assert recgen_filt.channel_positions.shape == (num_chan, 3)
                        assert recgen_filt.spike_traces.shape[1] == n_neurons
                        del recgen_filt

    def test_gen_recordings_drift(self):
        print("Test recording generation - drift")
        ne = 1
        ni = 1
        num_chan = self.num_chan
        n_neurons = ne + ni

        rec_params = mr.get_default_recordings_params()

        rec_params["spiketrains"]["n_exc"] = ne
        rec_params["spiketrains"]["n_inh"] = ni
        rec_params["spiketrains"]["duration"] = 5
        n_jitter = [1, 3]
        rec_params["recordings"]["drifting"] = True
        rec_params["recordings"]["slow_drift_velocity"] = 500
        rec_params["templates"]["min_dist"] = 1
        chunk_rec = [0, 2]

        modulations = ["none", "template", "electrode"]
        bursting = [False, True]
        drift_mode_speeds = ["slow", "fast"]
        drift_mode_probes = ["rigid", "non-rigid"]

        for i, mod in enumerate(modulations):
            for dms in drift_mode_speeds:
                for dmp in drift_mode_probes:
                    for b in bursting:
                        for j in n_jitter:
                            for ch in chunk_rec:
                                print(
                                    "Drifting: modulation",
                                    mod,
                                    "bursting",
                                    b,
                                    "jitter",
                                    j,
                                    "drift mode speed",
                                    dms,
                                    "drift mode probe",
                                    dmp,
                                    "chunk",
                                    ch,
                                )
                                rec_params["templates"]["n_jitters"] = j
                                rec_params["recordings"]["modulation"] = mod
                                rec_params["recordings"]["bursting"] = b
                                rec_params["recordings"]["chunk_duration"] = ch
                                rec_params["recordings"]["drift_mode_speed"] = dms
                                rec_params["recordings"]["drift_mode_probe"] = dmp

                                if i == len(modulations) - 1:
                                    rec_params["recordings"]["fs"] = 30000
                                    rec_params["recordings"]["n_drifting"] = 1
                                if mod == "electrode" and b is True and j == 5:
                                    rec_params["cell_types"] = None
                                    rec_params["recordings"]["shape_mod"] = True
                                recgen_drift = mr.gen_recordings(
                                    params=rec_params, tempgen=self.tempgen_drift, verbose=False
                                )
                                assert recgen_drift.recordings.shape[1] == num_chan
                                assert len(recgen_drift.spiketrains) == n_neurons
                                assert recgen_drift.channel_positions.shape == (num_chan, 3)
                                if j == 1:
                                    assert recgen_drift.templates.shape[0] == n_neurons
                                    assert recgen_drift.templates.shape[2] == num_chan
                                else:
                                    assert recgen_drift.templates.shape[0] == n_neurons
                                    assert recgen_drift.templates.shape[2] == j
                                    assert recgen_drift.templates.shape[3] == num_chan
                                assert recgen_drift.spike_traces.shape[1] == n_neurons
                                del recgen_drift

    def test_recording_custom_drifts(self):
        print("Test recording generation - drift")
        ne = 2
        ni = 1
        num_chan = self.num_chan
        n_neurons = ne + ni

        rec_params = mr.get_default_recordings_params()

        rec_params["spiketrains"]["n_exc"] = ne
        rec_params["spiketrains"]["n_inh"] = ni
        rec_params["spiketrains"]["duration"] = 10
        rec_params["recordings"]["drifting"] = True
        chunk_rec = [0, 2]

        # generate drift signals
        drift_dict1 = mr.get_default_drift_dict()
        drift_dict2 = mr.get_default_drift_dict()
        drift_dict3 = mr.get_default_drift_dict()

        drift_dict1["drift_mode_probe"] = "rigid"
        drift_dict1["drift_fs"] = 5
        drift_dict1["slow_drift_velocity"] = 10
        drift_dict1["slow_drift_amplitude"] = 50

        drift_dict2["drift_mode_speed"] = "fast"
        drift_dict2["fast_drift_period"] = 2
        drift_dict2["fast_drift_max_jump"] = 15

        drift_dict3["drift_mode_probe"] = "non-rigid"
        drift_dict3["drift_mode_speed"] = "slow"
        drift_dict3["slow_drift_waveform"] = "sine"
        drift_dict3["slow_drift_velocity"] = 80
        drift_dict3["slow_drift_amplitude"] = 10

        drift_dicts = [drift_dict1, drift_dict2, drift_dict3]

        for ch in chunk_rec:
            print("Drifting mixed: chunk", ch)
            rec_params["recordings"]["chunk_duration"] = ch

            recgen_drift = mr.gen_recordings(
                params=rec_params, tempgen=self.tempgen_drift, verbose=False, drift_dicts=drift_dicts
            )
            assert recgen_drift.recordings.shape[1] == num_chan
            assert len(recgen_drift.spiketrains) == n_neurons
            assert recgen_drift.channel_positions.shape == (num_chan, 3)
            assert recgen_drift.spike_traces.shape[1] == n_neurons
            assert len(recgen_drift.drift_list) == 3

            recgen_mixed_file = self.test_dir / "recordings_drift_mixed.h5"
            mr.save_recording_generator(recgen_drift, filename=recgen_mixed_file)
            recgen_drift_loaded = mr.load_recordings(recgen_mixed_file)
            assert len(recgen_drift_loaded.drift_list) == 3

            # test plotting drift
            _ = mr.plot_cell_drifts(recgen_drift_loaded)

            del recgen_drift_loaded
            del recgen_drift
            recgen_mixed_file.unlink()

    def test_save_load_templates(self):
        tempgen = mr.load_templates(self.test_dir / "templates.h5", verbose=True)
        tempgen_drift = mr.load_templates(self.test_dir / "templates_drift.h5")
        tempgen_drift_f = mr.load_templates(self.test_dir / "templates_drift.h5", return_h5_objects=True)

        assert np.allclose(tempgen.templates, self.tempgen.templates)
        assert np.allclose(tempgen.locations, self.tempgen.locations)
        assert np.allclose(tempgen.rotations, self.tempgen.rotations)
        assert np.allclose(tempgen_drift.templates, self.tempgen_drift.templates)
        assert np.allclose(tempgen_drift.locations, self.tempgen_drift.locations)
        assert np.allclose(tempgen_drift.rotations, self.tempgen_drift.rotations)
        assert np.allclose(tempgen_drift.templates, np.array(tempgen_drift_f.templates))

    def test_save_load_recordings(self):
        recgen_loaded = mr.load_recordings(
            self.test_dir / "recordings.h5",
            return_h5_objects=False,
            verbose=True,
        )
        recgen_loaded_f = mr.load_recordings(self.test_dir / "recordings.h5", return_h5_objects=False, verbose=True)
        recgen_loaded_r = mr.load_recordings(
            self.test_dir / "recordings.h5", load=["recordings"], return_h5_objects=True, verbose=True
        )

        assert np.allclose(recgen_loaded.templates, self.recgen.templates)
        assert np.allclose(recgen_loaded.recordings, self.recgen.recordings)
        assert np.allclose(recgen_loaded.spike_traces, self.recgen.spike_traces)
        assert np.allclose(recgen_loaded.voltage_peaks, self.recgen.voltage_peaks)
        assert np.allclose(recgen_loaded.channel_positions, self.recgen.channel_positions)
        assert np.allclose(recgen_loaded.timestamps.magnitude, self.recgen.timestamps.magnitude)
        assert np.allclose(recgen_loaded.recordings, np.array(recgen_loaded_f.recordings))
        assert np.allclose(recgen_loaded.recordings, np.array(recgen_loaded_r.recordings))

    def test_plots(self):
        _ = mr.plot_rasters(self.recgen.spiketrains)
        _ = mr.plot_rasters(self.recgen.spiketrains, overlap=True)
        _ = mr.plot_rasters(self.recgen.spiketrains, cell_type=True)
        _ = mr.plot_rasters(self.recgen.spiketrains, color="g")
        _ = mr.plot_rasters(self.recgen.spiketrains, color=["g"] * len(self.recgen.spiketrains))
        _ = mr.plot_recordings(self.recgen)
        _ = mr.plot_recordings(self.recgen, overlay_templates=True)
        _ = mr.plot_recordings(self.recgen, overlay_templates=True, max_channels_per_template=3)
        _ = mr.plot_templates(self.recgen)
        _ = mr.plot_templates(self.recgen, single_axes=True)
        _ = mr.plot_waveforms(self.recgen)
        _ = mr.plot_waveforms(self.recgen, electrode=0)
        _ = mr.plot_waveforms(self.recgen, electrode="max")
        _ = mr.plot_waveforms(self.recgen, electrode="max", max_waveforms=2)
        self.recgen.spiketrains[0].waveforms = None
        _ = mr.plot_waveforms(self.recgen, electrode="max")

    def test_extract_features(self):
        feat_t0 = mr.get_templates_features(
            self.tempgen.templates,
            feat_list=["neg", "pos", "amp", "width", "fwhm", "ratio", "speed"],
            dt=self.tempgen.info["params"]["dt"],
        )
        assert feat_t0["neg"].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0["pos"].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0["amp"].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0["width"].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0["fwhm"].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0["ratio"].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0["speed"].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])

    def test_recordings_resample(self):
        print("Test recording generation - resampling")
        ne = 2
        ni = 1
        num_chan = self.num_chan
        n_neurons = ne + ni
        fs = 10000
        duration = 5

        rec_params = mr.get_default_recordings_params()

        rec_params["spiketrains"]["n_exc"] = ne
        rec_params["spiketrains"]["n_inh"] = ni
        rec_params["spiketrains"]["duration"] = duration
        n_jitter = 3
        rec_params["templates"]["n_jitters"] = n_jitter
        rec_params["recordings"]["modulation"] = "none"
        rec_params["recordings"]["fs"] = fs
        recgen_rs = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

        assert recgen_rs.recordings.shape[0] == int(duration * fs)
        assert recgen_rs.recordings.shape[1] == num_chan
        assert recgen_rs.channel_positions.shape == (num_chan, 3)
        assert len(recgen_rs.spiketrains) == n_neurons
        assert len(recgen_rs.spiketrains) == n_neurons
        assert len(recgen_rs.spiketrains) == n_neurons
        assert recgen_rs.spike_traces.shape[1] == n_neurons
        del recgen_rs

    def test_recordings_backend(self):
        print("Test recording generation - backend")
        ne = 2
        ni = 1
        duration = 3

        rec_params = mr.get_default_recordings_params()

        rec_params["spiketrains"]["n_exc"] = ne
        rec_params["spiketrains"]["n_inh"] = ni
        rec_params["spiketrains"]["duration"] = duration
        n_jitter = 10
        rec_params["templates"]["n_jitters"] = n_jitter
        rec_params["recordings"]["modulation"] = "none"
        rec_params["recordings"]["filter"] = False

        rec_params["seeds"]["templates"] = 0
        rec_params["seeds"]["spiketrains"] = 0
        rec_params["seeds"]["convolution"] = 0
        rec_params["seeds"]["noise"] = 0

        n_jobs = [1, 2]
        chunk_durations = [0, 1]

        for n in n_jobs:
            for ch in chunk_durations:
                print("Test recording backend with", n, "jobs - chunk", ch)
                rec_params["recordings"]["chunk_duration"] = ch

                recgen_memmap = mr.gen_recordings(
                    params=rec_params, tempgen=self.tempgen, tmp_mode="memmap", verbose=False, n_jobs=n
                )
                recgen_np = mr.gen_recordings(
                    params=rec_params, tempgen=self.tempgen, tmp_mode=None, verbose=False, n_jobs=n
                )
                assert np.allclose(np.array(recgen_np.recordings), recgen_memmap.recordings.copy(), atol=1e-4)
                del recgen_memmap, recgen_np

    def test_recordings_seeds(self):
        print("Test recording generation - seeds")
        ne = 2
        ni = 1
        duration = 3

        rec_params = mr.get_default_recordings_params()

        rec_params["spiketrains"]["n_exc"] = ne
        rec_params["spiketrains"]["n_inh"] = ni
        rec_params["spiketrains"]["duration"] = duration
        n_jitter = 2
        rec_params["templates"]["n_jitters"] = n_jitter
        rec_params["recordings"]["modulation"] = "none"

        rec_params["seeds"]["templates"] = 0
        rec_params["seeds"]["spiketrains"] = 0
        rec_params["seeds"]["convolution"] = 0
        rec_params["seeds"]["noise"] = 0

        n_jobs = [1, 2]
        chunk_durations = [0, 1]

        for n in n_jobs:
            for ch in chunk_durations:
                print("Test recording seeds with", n, "jobs - chunk", ch)
                rec_params["chunk_duration"] = n

                print("memmap")
                recgen1 = mr.gen_recordings(
                    params=rec_params, tempgen=self.tempgen, tmp_mode="memmap", verbose=False, n_jobs=n
                )
                recgen2 = mr.gen_recordings(
                    params=rec_params, tempgen=self.tempgen, tmp_mode="memmap", verbose=False, n_jobs=n
                )

                assert np.allclose(np.array(recgen1.recordings), np.array(recgen2.recordings), atol=1e-4)
                del recgen1, recgen2

                print("memory")
                recgen1 = mr.gen_recordings(
                    params=rec_params, tempgen=self.tempgen, tmp_mode=None, verbose=False, n_jobs=n
                )
                recgen2 = mr.gen_recordings(
                    params=rec_params, tempgen=self.tempgen, tmp_mode=None, verbose=False, n_jobs=n
                )

                assert np.allclose(np.array(recgen1.recordings), recgen2.recordings, atol=1e-4)
                del recgen1, recgen2

    def test_recordings_dtype(self):
        print("Test recording generation - dtype")
        ne = 2
        ni = 1
        duration = 1

        dtypes = ["int16", "int32", "float16", "float32", "float64"]
        modulations = ["none", "template", "electrode"]

        rec_params = mr.get_default_recordings_params()
        rec_params["spiketrains"]["n_exc"] = ne
        rec_params["spiketrains"]["n_inh"] = ni
        rec_params["spiketrains"]["duration"] = duration
        n_jitter = 3
        rec_params["templates"]["n_jitters"] = n_jitter

        for i, dt in enumerate(dtypes):
            for mod in modulations:
                rec_params["recordings"]["modulation"] = mod

                print("Dtype:", dt, "modulation", mod)
                rec_params["recordings"]["dtype"] = dt
                recgen_dt = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

                assert recgen_dt.recordings[0, 0].dtype == dt
                del recgen_dt

    def test_adc_bit_depth_lsb_gain(self):
        print("Test recording generation - adc depth and lsb")
        ne = 2
        ni = 1
        duration = 1

        bit_depths = [10, 11, 12]
        lsbs = [1, 2, 4]
        gains = [0.1, 0.2, 0.3]

        rec_params = mr.get_default_recordings_params()
        rec_params["spiketrains"]["n_exc"] = ne
        rec_params["spiketrains"]["n_inh"] = ni
        rec_params["spiketrains"]["duration"] = duration
        n_jitter = 3
        rec_params["templates"]["n_jitters"] = n_jitter
        rec_params["recordings"]["dtype"] = "int16"
        rec_params["recordings"]["filter"] = False

        for bd in bit_depths:
            for lsb in lsbs:
                rec_params["recordings"]["adc_bit_depth"] = bd
                rec_params["recordings"]["lsb"] = lsb

                print("ADC bit depth:", bd, "lsb", lsb)
                recgen_adc_lsb = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

                assert np.ptp(recgen_adc_lsb.recordings) / lsb <= 2**bd
                lsb_rec = np.min(np.diff(np.sort(np.unique(recgen_adc_lsb.recordings.ravel()))))
                assert lsb_rec == lsb

                # save and reload gain
                save_path = self.test_dir / f"{bd}_{lsb}.h5"
                mr.save_recording_generator(recgen_adc_lsb, save_path)
                recgen_loaded = mr.load_recordings(save_path)
                assert recgen_loaded.gain_to_uV == recgen_adc_lsb.gain_to_uV

                del recgen_adc_lsb
                save_path.unlink()

        for gain in gains:
            for lsb in lsbs:
                rec_params["recordings"]["gain"] = gain
                rec_params["recordings"]["lsb"] = lsb

                print("Gain:", gain, "lsb", lsb)
                recgen_gain_lsb = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

                assert recgen_gain_lsb.gain_to_uV == gain
                lsb_rec = np.min(np.diff(np.sort(np.unique(recgen_gain_lsb.recordings.ravel()))))
                assert lsb_rec == lsb

                # save and reload gain
                save_path = self.test_dir / f"{gain}_{lsb}.h5"
                mr.save_recording_generator(recgen_gain_lsb, save_path)
                recgen_loaded = mr.load_recordings(save_path)
                assert recgen_loaded.gain_to_uV == recgen_gain_lsb.gain_to_uV

                del recgen_gain_lsb
                save_path.unlink()

    def test_default_params(self):
        print("Test default params")
        info, info_folder = mr.get_default_config()
        cell_models_folder = info["cell_models_folder"]
        tempgen = mr.gen_templates(cell_models_folder, params={"n": 2}, templates_tmp_folder=info["templates_folder"])
        recgen = mr.gen_recordings(templates=self.test_dir / "templates.h5", verbose=False)
        recgen.params["recordings"]["noise_level"] = 0
        recgen.generate_recordings()
        recgen_loaded = mr.load_recordings(self.test_dir / "recordings.h5", verbose=True)
        recgen_loaded.params["recordings"]["noise_level"] = 0
        recgen_loaded.generate_recordings()
        recgen_empty = mr.RecordingGenerator(rec_dict={}, info={})

        n = 2
        num_cells = self.num_cells
        templates_params = self.templates_params

        assert tempgen.templates.shape[0] == (n * num_cells)
        assert len(tempgen.locations) == (n * num_cells)
        assert len(tempgen.rotations) == (n * num_cells)
        assert len(tempgen.celltypes) == (n * num_cells)
        assert len(np.unique(tempgen.celltypes)) == num_cells
        assert (
            np.min(tempgen.locations[:, 0]) > templates_params["xlim"][0]
            and np.max(tempgen.locations[:, 0]) < templates_params["xlim"][1]
        )

        assert recgen.recordings.shape[1] == self.num_chan
        assert recgen.channel_positions.shape == (self.num_chan, 3)
        assert recgen_loaded.recordings.shape[1] == self.num_chan
        assert recgen_loaded.channel_positions.shape == (self.num_chan, 3)
        assert len(recgen_empty.recordings) == 0
        del recgen, recgen_empty

    def test_cli(self):
        default_config, mearec_home = mr.get_default_config()

        runner = CliRunner()
        result = runner.invoke(cli, ["--help"])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["default-config"])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["available-probes"])
        assert result.exit_code == 0
        result = runner.invoke(
            cli,
            [
                "gen-templates",
                "-n",
                "2",
                "--no-parallel",
                "--recompile",
                "-r",
                "3drot",
                "-nc",
                "2",
                "-ov",
                "20",
                "-s",
                "1",
                "-mind",
                "10",
                "-maxd",
                "100",
                "-drst",
                "10",
                "-v",
            ],
        )
        assert result.exit_code == 0
        result = runner.invoke(
            cli,
            [
                "gen-recordings",
                "-t",
                str(self.test_dir / "templates.h5"),
                "-ne",
                "2",
                "-ni",
                "1",
                "-fe",
                "5",
                "-fi",
                "15",
                "-se",
                "1",
                "-si",
                "1",
                "-mr",
                "0.2",
                "-rp",
                "2",
                "-p",
                "poisson",
                "-md",
                "1",
                "-mina",
                "10",
                "-maxa",
                "1000",
                "--fs",
                "32000",
                "-sr",
                "0",
                "-sj",
                "1",
                "-nl",
                "10",
                "-m",
                "none",
                "-chd",
                "0",
                "-nseed",
                "10",
                "-hd",
                "30",
                "-cn",
                "-cp",
                "500",
                "-cq",
                "1",
                "-rnf",
                "1",
                "-stseed",
                "100",
                "-tseed",
                "10",
                "--filter",
                "-fc",
                "500",
                "-fo",
                "3",
                "--overlap",
                "-ot",
                "0.8",
                "--extract-wf",
                "-angt",
                "15",
                "-drvel",
                "10",
                "-tsd",
                "1",
                "-v",
            ],
        )
        print(result.output)
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-params", "."])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-params", default_config["templates_params"]])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-params", "."])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-params", default_config["recordings_params"]])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-cell-models-folder", "./cell_models"])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-cell-models-folder", default_config["cell_models_folder"]])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-folder", "."])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-folder", "."])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-folder", "./templates", "--create"])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-folder", "./recordings", "--create"])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-folder", default_config["templates_folder"]])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-folder", default_config["recordings_folder"]])
        assert result.exit_code == 0

    def test_simulate_cell(self):
        cell_folder = Path(mr.get_default_cell_models_folder())
        params = mr.get_default_templates_params()

        target_spikes = [3, 50]
        params["target_spikes"] = target_spikes
        cell_path = [c for c in cell_folder.iterdir() if "TTPC1" in c.name][0]
        cell_name = cell_path.parts[-1]

        cell, v, i = mr.run_cell_model(
            cell_model_folder=str(cell_path), sim_folder=None, verbose=True, save=False, return_vi=True, **params
        )
        c = mr.return_bbp_cell_morphology(str(cell_name), cell_folder)
        assert target_spikes[0] <= len(v) <= target_spikes[1]
        assert target_spikes[0] <= len(i) <= target_spikes[1]
        assert len(c.x) == len(c.y) and len(c.x) == len(c.z)


if __name__ == "__main__":
    test = TestGenerators()
    test.setUpClass()
    # TestGenerators().test_gen_recordings_drift()
    test.test_gen_recordings_filters()
    # test.test_simulate_cell()



import os
import shutil
import sys
import time
from copy import deepcopy
from pathlib import Path

import MEAutility as mu
import numpy as np
import yaml
from joblib import Parallel, cpu_count, delayed
from packaging.version import parse

from ..simulate_cells import (compute_eap_based_on_tempgen,
                              compute_eap_for_cell_model)
from ..tools import (clean_dict_for_yaml, get_default_config, load_tmp_eap,
                     safe_yaml_load)

_intra_keys = ["sim_time", "target_spikes", "cut_out", "dt", "delay", "weights", "seed", "cell_models_folder"]


def simulate_cell_templates(i, simulate_script, tot, cell_model, model_folder, intraonly, params_path, verbose):
    model_folder = Path(model_folder)
    print(f"Starting simulation {i + 1}/{tot} - cell: {Path(cell_model).name}\n", flush=True)
    python = sys.executable
    if verbose:
        verbose = 1
    else:
        verbose = 0
    cmd = (
        f"{python} {simulate_script} {i} {str(model_folder / cell_model)} "
        f"{intraonly} {params_path.absolute()} {verbose}"
    )
    os.system(cmd)


class TemplateGenerator:
    """
    Class for generation of templates called by the gen_templates function.
    The list of parameters is in default_params/templates_params.yaml.

    Parameters
    ----------
    cell_models_folder : str
        Path to folder containing Blue Brain Project cell models
    templates_folder : str
        Path to output template folder (if not in params)
    temp_dict :  dict
        Dictionary to instantiate TemplateGenerator with existing data. It contains the following fields:
          - templates : float (n_templates, n_electrodes, n_timepoints)
          - locations : float (n_templates, 3)
          - rotations : float (n_templates, 3)
          - celltypes : str (n_templates)
    info :  dict
        Info dictionary to instantiate TemplateGenerator with existing data. It contains the following fields:
          - params : dict with template generation parameters
          - electrodes : dict with probe info (from MEAutility.return_mea_info('probe-name'))
    tempgen : TemplateGenerator
        If a TemplateGenerator is passed, the cell types, locations, and rotations of the templates will be set using
        the provided templates
    params : dict
        Dictionary with parameters to simulate templates. Default values can be retrieved with
        mr.get_default_template_params()
    intraonly : bool
        If True, only intracellular simulations are performed
    parallel : bool
        If True, cell models are run in parallel
    recompile: bool
        If True, cell models are recompiled (suggested if new models are added)
    n_jobs: int
        If None, all cpus are used
    delete_tmp : bool
        If True, temporary files are removed
    verbose : bool
        If True, output is verbose
    """

    def __init__(
        self,
        cell_models_folder=None,
        templates_folder=None,
        temp_dict=None,
        info=None,
        tempgen=None,
        params=None,
        intraonly=False,
        parallel=True,
        recompile=False,
        n_jobs=None,
        joblib_backend="loky",
        delete_tmp=True,
        verbose=False,
    ):
        self._verbose = verbose
        if temp_dict is not None and info is not None:
            if "templates" in temp_dict.keys():
                self.templates = temp_dict["templates"]
            if "locations" in temp_dict.keys():
                self.locations = temp_dict["locations"]
            if "rotations" in temp_dict.keys():
                self.rotations = temp_dict["rotations"]
            if "celltypes" in temp_dict.keys():
                self.celltypes = temp_dict["celltypes"]
            self.info = info
            self.params = deepcopy(info)
        else:
            if cell_models_folder is None:
                raise AttributeError("Specify cell folder!")
            if params is None:
                if self._verbose:
                    print("Using default parameters")
                self.params = {}
            else:
                self.params = deepcopy(params)
            self.cell_model_folder = Path(cell_models_folder).resolve()
            self.n_jobs = n_jobs
            self.joblib_backend = joblib_backend
            if templates_folder is not None:
                templates_folder = Path(templates_folder).resolve()
            self.templates_folder = templates_folder
            self.tempgen = tempgen
            self.simulation_params = {
                "intraonly": intraonly,
                "parallel": parallel,
                "delete_tmp": delete_tmp,
                "recompile": recompile,
            }

    def generate_templates(self):
        """
        Generate templates.
        """
        cell_models_folder = self.cell_model_folder
        templates_folder = self.templates_folder
        intraonly = self.simulation_params["intraonly"]
        parallel = self.simulation_params["parallel"]
        recompile = self.simulation_params["recompile"]
        delete_tmp = self.simulation_params["delete_tmp"]

        if cell_models_folder.is_dir():
            cell_models = [
                f for f in cell_models_folder.iterdir() if "mods" not in f.name and not f.name.startswith(".")
            ]
            if len(cell_models) == 0:
                raise AttributeError(cell_models_folder, " contains no cell models!")
        else:
            raise NotADirectoryError("Cell models folder: does not exist!")

        this_dir, this_filename = os.path.split(__file__)
        simulate_script = str(Path(this_dir).parent / "simulate_cells.py")

        # Compile NEURON models (nrnivmodl)
        if not (cell_models_folder / "mods").is_dir() or recompile:
            if self._verbose:
                print("Compiling NEURON models")
            python = sys.executable
            os.system(f"{python} {simulate_script} compile {cell_models_folder}")

        # sort cell model names
        cell_models = np.array(cell_models)[np.argsort([f.name for f in cell_models])]

        if "sim_time" not in self.params.keys():
            self.params["sim_time"] = 1
        if "target_spikes" not in self.params.keys():
            self.params["target_spikes"] = [3, 50]
        if "cut_out" not in self.params.keys():
            self.params["cut_out"] = [2, 5]
        if "dt" not in self.params.keys():
            self.params["dt"] = 2**-5
        if "delay" not in self.params.keys():
            self.params["delay"] = 10
        if "weights" not in self.params.keys():
            self.params["weights"] = [0.25, 1.75]

        if "rot" not in self.params.keys():
            self.params["rot"] = "physrot"
        if "probe" not in self.params.keys():
            available_mea = mu.return_mea_list()
            probe = available_mea[np.random.randint(len(available_mea))]
            if self._verbose:
                print("Probe randomly set to: %s" % probe)
            self.params["probe"] = probe
        if "ncontacts" not in self.params.keys():
            self.params["ncontacts"] = 1
        if "overhang" not in self.params.keys():
            self.params["overhang"] = 1
        if "xlim" not in self.params.keys():
            self.params["xlim"] = [10, 80]
        if "ylim" not in self.params.keys():
            self.params["ylim"] = None
        if "zlim" not in self.params.keys():
            self.params["zlim"] = None
        if "x_distr" not in self.params.keys():
            self.params["x_distr"] = "uniform"
        if "beta_distr_params" not in self.params.keys():
            self.params["beta_distr_params"] = [1.5, 5]
        if "offset" not in self.params.keys():
            self.params["offset"] = 0
        if "det_thresh" not in self.params.keys():
            self.params["det_thresh"] = 30
        if "n" not in self.params.keys():
            self.params["n"] = 50
        if "check_eap_shape" not in self.params.keys():
            self.params["check_eap_shape"] = True
        if "min_amp" not in self.params.keys():
            self.params["min_amp"] = 30
        if "seed" not in self.params.keys():
            self.params["seed"] = np.random.randint(1, 10000)
        elif self.params["seed"] is None:
            self.params["seed"] = np.random.randint(1, 10000)
        if templates_folder is None:
            info, _ = get_default_config()
            self.params["templates_folder"] = info["templates_folder"]
            templates_folder = Path(self.params["templates_folder"])
        else:
            self.params["templates_folder"] = str(templates_folder)
        self.params["cell_models_folder"] = str(cell_models_folder)
        if "drifting" not in self.params.keys():
            self.params["drifting"] = False
        if "max_drift" not in self.params.keys():
            self.params["max_drift"] = 100
        if "min_drift" not in self.params.keys():
            self.params["min_drift"] = 30
        if "drift_steps" not in self.params.keys():
            self.params["drift_steps"] = 10
        if "drift_xlim" not in self.params.keys():
            self.params["drift_xlim"] = [-10, 10]
        if "drift_ylim" not in self.params.keys():
            self.params["drift_ylim"] = [-10, 10]
        if "drift_zlim" not in self.params.keys():
            self.params["drift_zlim"] = [20, 80]
        if "check_for_drift_amp" not in self.params.keys():
            self.params["check_for_drift_amp"] = False
        if "drift_within_bounds" not in self.params.keys():
            self.params["drift_within_bounds"] = False

        rot = self.params["rot"]
        n = self.params["n"]
        probe = self.params["probe"]

        # check intra params
        intra_params = {k: v for k, v in self.params.items() if k in _intra_keys}
        # check params
        intracellular_folder = Path(self.params["templates_folder"]) / "intracellular"
        skip_existing_intracellular = check_intracellular_params(intracellular_folder, intra_params)
        if skip_existing_intracellular:
            if intracellular_folder.is_dir():
                if self._verbose:
                    print(f"Removing intracellular folder {intracellular_folder} because of intra parameter mismatch")
                shutil.rmtree(intracellular_folder)

        tmp_params_path = Path("tmp_params_path.yaml")
        with open(tmp_params_path, "w") as f:
            # alessio we did have bug here because some params are numpy.int, numpy.bool
            # I did this fast debug but we need a way to convert then to standard float/int/bool
            # yaml.dump(self.params, f)
            yaml.dump(clean_dict_for_yaml(self.params), f)

        if self.tempgen is not None and parallel and self.n_jobs not in (0, 1):
            print(
                "\nWARNING: Generation of templates from a template generator is only supported without parallel "
                "processing. Setting parallel to False\n"
            )
            parallel = False

        # Simulate neurons and EAP for different cell models separately
        if parallel and self.n_jobs not in (0, 1):
            start_time = time.time()
            tot = len(cell_models)
            if self.n_jobs is None:
                n_jobs = cpu_count()
                print(f"Setting n_jobs to {n_jobs} CPUs")
            else:
                n_jobs = self.n_jobs

            if self._verbose:
                print("Running with", n_jobs, "jobs")

            Parallel(n_jobs=n_jobs, backend=self.joblib_backend)(
                delayed(simulate_cell_templates)(
                    i,
                    simulate_script,
                    tot,
                    cell_model,
                    cell_models_folder,
                    intraonly,
                    tmp_params_path,
                    self._verbose,
                )
                for i, cell_model in enumerate(cell_models)
            )
        else:
            start_time = time.time()
            if self.tempgen is None:
                for i, cell_model in enumerate(cell_models):
                    if self._verbose:
                        print(f"\n\n {cell_model} {i + 1}/{len(cell_models)}\n\n")
                    compute_eap_for_cell_model(
                        i,
                        cell_model=cell_model,
                        params_path=tmp_params_path,
                        intraonly=intraonly,
                        verbose=self._verbose,
                    )

            else:
                print("Using template generation info")
                compute_eap_based_on_tempgen(
                    cell_folder=cell_models_folder,
                    params_path=tmp_params_path,
                    tempgen=self.tempgen,
                    intraonly=intraonly,
                    verbose=self._verbose,
                )
        # save new intracellular params
        if skip_existing_intracellular:
            params_file = intracellular_folder / "intra_params.yaml"
            if params_file.is_file():
                params_file.unlink()
            with params_file.open("w") as f:
                yaml.dump(intra_params, f)
            if self._verbose:
                print(f"Saving new intracellular parameters in {params_file}")

        tmp_folder = Path(templates_folder) / rot / f"tmp_{n}_{probe}"
        if not Path(tmp_folder).is_dir():
            raise FileNotFoundError(f"{tmp_folder} not found. Something went wrong in the template generation phase.")

        print("Aggregating templates")
        templates, locations, rotations, celltypes = load_tmp_eap(tmp_folder)
        if delete_tmp:
            shutil.rmtree(tmp_folder)
            os.remove(tmp_params_path)

        self.info = {}

        self.templates = templates
        self.locations = locations
        self.rotations = rotations
        self.celltypes = celltypes

        self.info["params"] = self.params
        self.info["electrodes"] = mu.return_mea_info(probe)

        print(f"\n\n\nSimulation time: {time.time() - start_time}\n\n\n")


def check_intracellular_params(
    vm_im_sim_folder, params, check_params=["dt", "cut_out", "cell_models_folder", "target_spikes"]
):
    skip_existing_intracellular = False
    if not vm_im_sim_folder.is_dir():
        skip_existing_intracellular = True
    else:
        params_files = [f for f in Path(vm_im_sim_folder).iterdir() if "intra_params.yaml" in f.name]
        if len(params_files) == 0:
            skip_existing_intracellular = True
        if len(params_files) == 1:
            params_file = params_files[0]
            existing_intra_params = safe_yaml_load(params_file)
            for param_key in check_params:
                if existing_intra_params[param_key] != params[param_key]:
                    print(f"{param_key} is different!")
                    skip_existing_intracellular = True
    return skip_existing_intracellular
